
import urllib.parse as urlparse
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from pydantic import BaseModel, Field, PrivateAttr

from redbird.oper import GreaterEqual, GreaterThan, LessEqual, LessThan, NotEqual, Operation
from redbird.base import BaseResult, BaseRepo



from redbird.templates import TemplateRepo

try:
    import requests
except ImportError:
    # Cannot import, we cannot create request Session
    # RESTRepo cannot be used
    pass
else:
    class Session(requests.Session):
        """Subclassed requests.Session for more
        unified methods.
        """

        def remove(self):
            self.close()

class RESTRepo(TemplateRepo):

    """REST API Repository

    Parameters
    ----------
    url : str
        Base URL for the API. Should not contain
        query parameters or the item ID. The full
        URL will be "{url}/{id}" if a single item
        is searched, "{url}?{param}={value}" if 
        the endpoint is queried.
    model : Type
        Class of an item in the repository.
        Commonly dict or subclass of Pydantic
        BaseModel. By default dict
    id_field : str, optional
        Attribute or key that identifies each item
        in the repository.
    field_access : {'attr', 'key'}, optional
        How to access a field in an item. Either
        by attribute ('attr') or key ('item').
        By default guessed from the model.
    query : Type, optional
        Query model of the repository.
    errors_query : {'raise', 'warn', 'discard'}
        Whether to raise an exception, warn or discard
        the item in case of validation error in 
        converting data to the item model from
        the repository. By default raise
    headers : dict
        HTTP Headers (for example authentication)
    url_params : dict
        Additional query parameters passed to every 
        request.
    result : str, callable
        Where the list of items is found from the output.

    Examples
    --------

    .. code-block:: python

        repo = MemoryRepo(url="http://example.com/api")

    For services requiring authentication via headers:
    .. code-block:: python

        token = "1234567890"
        repo = RESTRepo(url="http://example.com/api", headers={"Authorization": f"Bearer {token}"})

    You may also supply some URL parameters for every query:

    .. code-block:: python

        repo = RESTRepo(url="http://example.com/api", url_params={"fields": "car_type,car_model,registration_number"})
    """
    result: Optional[Union[str, Callable]]
    url: str
    url_params: dict = {}
    headers: dict = {}

    _session = PrivateAttr()

    ordered: bool = Field(default=False, const=True)

    def insert(self, item):
        json = self.item_to_dict(item)
        self._request(
            "POST",
            self.url,
            json=json
        )

    def replace(self, item):
        qry = {self.id_field: self.get_field_value(item, self.id_field)}
        values = self.item_to_dict(item)
        self.filter_by(**qry).replace(**values)

    def _request(self, *args, **kwargs):
        page = self.session.request(
            *args, **kwargs, headers=self.headers
        )
        return page

    def query_data(self, query):
        url = query
        page = self._request("GET", url)
        output = page.json()

        result_loc = self.result
        if result_loc is None:
            items = output
        elif isinstance(result_loc, str):
            items = output[result_loc]
        elif isinstance(result_loc, callable):
            items = result_loc(output)
        else:
            raise TypeError(f"Could not locate results from {result_loc}")

        if isinstance(items, (list, tuple, set)):
            for data in items:
                yield data
        else:
            yield items

    def query_delete(self, query):
        url = query
        self._request("DELETE", url)

    def query_update(self, query, values):
        url = query
        self._request("PATCH", url, json=values)

    def query_replace(self, query, values):
        url = query
        self._request("PUT", url, json=values)

    def format_query(self, query:dict) -> str:
        "Turn the query to a form that's understandable by the underlying API"
        id = query.pop(self.id_field) if query is not None and self.id_field in query else None

        url_base = self.url
        url_params = self._get_url_params(query)

        # URL should look like "www.example.com/api/items/{id}?{param}={value}"
        # or "www.example.com/api/items/{id}"
        # or "www.example.com/api/items?{param}={value}"
        # or "www.example.com/api/items"
        if id and url_params:
            return f"{url_base}/{id}?{url_params}"
        elif id:
            return f"{url_base}/{id}"
        elif url_params:
            return f"{url_base}?{url_params}"
        else:
            return url_base

    def _get_url_params(self, query:dict) -> str:
        query = {} if query is None else query
        url_params = self.url_params.copy()
        url_params.update(query)
        return urlparse.urlencode(url_params)

    @property
    def session(self):
        if not hasattr(self, "_session"):
            self._session = Session()
        return self._session
        



from functools import wraps
from operator import getitem
from typing import Dict, Generator, List
import warnings

from redbird.base import Data, Item
from redbird.exc import ConversionWarning, _handle_conversion_error
from redbird.templates import TemplateRepo
from redbird.oper import Operation

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class QueryMatcher:

    def __init__(self, query, value_getter):
        self.query = query
        self.value_getter = value_getter

    def __contains__(self, item):
        "Match whether item fulfills the query"
        return all(
            val.evaluate(self.get_value(item, key)) if isinstance(val, Operation) else self.get_value(item, key) == val
            for key, val in self.query.items()
        )

    def get_value(self, item, key):
        try:
            return self.value_getter(item, key)
        except KeyError:
            return None

def read_items(repo: TemplateRepo, reader: Generator, query: dict) -> List[Item]:
    
    query = QueryMatcher(query, value_getter=repo.get_field_value)
    for data in reader:
        try:
            item = repo.data_to_item(data)
        except ValueError:
            _handle_conversion_error(repo, data)
        else:
            if item in query:
                yield item

def update_items(repo: TemplateRepo, reader: Generator, query:dict, values:Dict, return_:Literal['all', 'updated']='all') -> List[Item]:
    query = QueryMatcher(query, value_getter=repo.get_field_value)
    for data in reader:
        item = repo.data_to_item(data)
        is_updated = item in query
        if is_updated:
            for key, val in values.items():
                repo.set_field_value(item, key, val)
        
        if return_ == 'all':
            yield item
        elif return_ == 'updated' and is_updated:
            yield item

def delete_items(repo: TemplateRepo, reader: Generator, query:dict, return_:Literal['remained', 'deleted']='remained') -> List[Item]:
    query = QueryMatcher(query, value_getter=repo.get_field_value)
    for data in reader:
        item = repo.data_to_item(data)
        is_deleted = item in query

        if is_deleted and return_ == 'deleted':
            yield item
        elif not is_deleted and return_ == 'remained':
            yield item




import re

def to_camelcase(string):
    "snake_case --> camelCase"
    # TODO: Split using regex (with also space)
    if is_camel_case(string):
        return string
    elif is_pascal_case(string) and not string.isupper():
        return string[0].lower() + string[1:]
    words = re.sub(r"[^A-Za-z0-9]+", "_", string).split("_")
    return ''.join(
        [
            word.capitalize() if i > 0 else word.lower()
            for i, word in enumerate(
                words
            )
        ]
    )


def to_snakecase(string):
    "--> snake_case"
    if is_snake_case(string):
        # Examples: snake_case
        return string
    elif string.isupper():
        # Examples: TITLE
        return string.lower()
    if is_camel_case(string) or is_pascal_case(string):
        # Put "_" before each capitalized letter (if not preceded with one like "TML" in "HTML")
        string = re.sub(r'(?!^)([A-Z][A-Z]*)', r'_\1', string)
    else:
        string = re.sub(r"[^A-Za-z0-9]+", "_", string)
    return (
        string
        .lower()
    )

def is_camel_case(string):
    return bool(re.match(r"^[a-z]+([A-Z][a-z]*)*$", string))

def is_snake_case(string):
    return bool(re.match(r"^[a-z_][0-9a-z_]*$", string))

def is_pascal_case(string):
    return bool(re.match(r"^([A-Z][a-z]*)+$", string))

def to_case(value:str, case:str):
    func = {
        "snake": to_snakecase,
        "camel": to_camelcase,
    }[case]
    return func(value)



from abc import abstractmethod
import warnings
from typing import Any, ClassVar, Dict, Iterator, List, Union

from redbird.exc import ConversionWarning, _handle_conversion_error
from .base import BaseRepo, BaseResult, Data, Item

class TemplateResult(BaseResult):
    repo: 'TemplateRepo'

    def query(self) -> Iterator[Item]:
        "Get actual result"
        try:
            items = self.repo.query_items(self.query_)
        except NotImplementedError:
            for data in self.repo.query_data(self.query_):
                try:
                    yield self.repo.data_to_item(data)
                except ValueError:
                    _handle_conversion_error(self.repo, data)
        else:
             yield from items

    def query_data(self) -> Iterator[Data]:
        "Get actual result"
        return self.repo.query_data(self.query_)

    def update(self, **kwargs):
        return self.repo.query_update(self.query_, kwargs)

    def delete(self):
        return self.repo.query_delete(self.query_)

    def replace(self, __values:dict=None, **kwargs):
        "Replace the existing item(s) with given"
        if __values is not None:
            kwargs.update(__values)
        try:
            return self.repo.query_replace(self.query_, kwargs)
        except NotImplementedError:
            # Has no custom replace implemented, using 
            # default alternative (delete, add)
            return super().replace(**kwargs)

    def count(self) -> int:
        try:
            return self.repo.query_count(self.query_)
        except NotImplementedError:
            # Has no custom replace implemented, using 
            # default alternative (len(self))
            return super().count()

    def first(self) -> Item:
        try:
            return self.repo.query_read_first(self.query_)
        except NotImplementedError:
            return super().first()

    def limit(self, n:int) -> List[Item]:
        try:
            return self.repo.query_read_limit(self.query_, n)
        except NotImplementedError:
            return super().limit(n)

    def last(self) -> Item:
        try:
            return self.repo.query_read_last(self.query_)
        except NotImplementedError:
            return super().last()

    def format_query(self, query: dict) -> dict:
        qry = super().format_query(query)
        try:
            return self.repo.format_query(qry)
        except NotImplementedError:
            return qry


class TemplateRepo(BaseRepo):
    """Template repository for easy subclassing

    Parameters
    ----------
    model : Type
        Class of an item in the repository.
        Commonly dict or subclass of Pydantic
        BaseModel. By default dict
    id_field : str, optional
        Attribute or key that identifies each item
        in the repository.
    field_access : {'attr', 'key'}, optional
        How to access a field in an item. Either
        by attribute ('attr') or key ('item').
        By default guessed from the model.
    query : Type, optional
        Query model of the repository.
    errors_query : {'raise', 'warn', 'discard'}
        Whether to raise an exception, warn or discard
        the item inn case of validation error in 
        converting data to the item model from
        the repository. By default raise 

    Examples
    --------

    .. code-block:: python

        class MyRepo(TemplateRepo):

            def insert(self, item):
                # Insert item to the data store
                ...

            def query_read(self, query: dict):
                # Get data from repository
                for item in ...: 
                    yield item

            def query_update(self, query: dict, values: dict):
                # Update items with values matcing the query
                ...

            def query_delete(self, query: dict):
                # Delete items matcing the query
                ...

            def item_to_data(self, item):
                # Convert item to type that is understandable
                # by the repository's data store
                ...
                return data
    """
    cls_result: ClassVar = TemplateResult

    def query_data(self, query: dict) -> Iterator[Data]:
        """Query (read) the data store and return raw data

        Override this or :func:`~redbird.templates.TemplateRepo.query_items` method.

        Parameters
        ----------
        query : dict
            Repository query, by default dict.

        Returns
        -------
        iterable (any)
            Iterable of raw data that is converted to an item using :func:`~redbird.base.BaseRepo.data_to_item`

        Examples
        --------

        Used in following cases:

        .. code-block:: python

            repo.filter_by(color="red").all()
        
        """
        raise NotImplementedError("Method query_data not implemented")

    def query_items(self, query: dict) -> Iterator[Item]:
        """Query (read) the data store and return items

        Override this or :func:`~redbird.templates.TemplateRepo.query_data` method.

        Parameters
        ----------
        query : dict
            Repository query, by default dict.

        Returns
        -------
        iterable (``model``)
            Items that are instances of the class in the ``model`` attibute.
            Typically dicts or instances of subclassed Pydantic BaseModel

        Examples
        --------

        Used in following cases:

        .. code-block:: python

            repo.filter_by(color="red").all()
        
        """
        raise NotImplementedError("Method query_items not implemented")

    @abstractmethod
    def query_update(self, query: dict, values: dict):
        """Update the result of the query with given values

        Override this method.

        Parameters
        ----------
        query : any
            Repository query, by default dict.

        Examples
        --------

        Used in following cases:

        .. code-block:: python

            repo.filter_by(color="red").update(color="blue")

        """
        ...

    @abstractmethod
    def query_delete(self, query: dict):
        """Delete the result of the query

        Override this method.

        Parameters
        ----------
        query : any
            Repository query, by default dict.

        Examples
        --------

        Used in following cases:

        .. code-block:: python

            repo.filter_by(color="red").delete()

        """
        ...

    def query_read_first(self, query: dict) -> Item:
        """Query the first item

        You may override this method. By default,
        the first item returned by :class:`TemplateRepo.query_read`
        is returned.

        Parameters
        ----------
        query : any
            Repository query, by default dict.

        Examples
        --------

        Used in the following case:

        .. code-block:: python

            repo.filter_by(color="red").first()

        """
        raise NotImplementedError("Read using first not implemented.")

    def query_read_limit(self, query: dict, n:int) -> List[Item]:
        """Query the first n items

        You may override this method. By default,
        the N first items returned by :class:`TemplateRepo.query_read`
        are returned.

        Parameters
        ----------
        query : any
            Repository query, by default dict.
        n : int
            Number of items to return

        Examples
        --------

        Used in the following case:

        .. code-block:: python

            repo.filter_by(color="red").limit(3)

        """
        raise NotImplementedError("Read using limit not implemented.")

    def query_read_last(self, query: dict) -> Item:
        """Query the last item

        You may override this method. By default,
        the last item returned by :class:`TemplateRepo.query_read`
        is returned.

        Parameters
        ----------
        query : any
            Repository query, by default dict.

        Examples
        --------

        Used in the following case:

        .. code-block:: python

            repo.filter_by(color="red").last()

        """
        raise NotImplementedError("Read using first not implemented.")

    def query_replace(self, query: dict, values: dict):
        """Replace the items with given values using given query

        You may override this method. By default,
        the result of the query is deleted and an
        item from the values is generated.

        Parameters
        ----------
        query : any
            Repository query, by default dict.
        values : dict
            Values to replace the items' existing
            values with

        Examples
        --------

        Used in the following case:

        .. code-block:: python

            repo.filter_by(color="red").replace(color="blue")

        """
        raise NotImplementedError("Replacing using query not implemented.")

    def query_count(self, query: dict):
        """Count the items returned by the query

        You may override this method. By default,
        the items returned by :class:`TemplateRepo.query_read`
        are counted.

        Parameters
        ----------
        query : any
            Repository query, by default dict.

        Examples
        --------

        Used in the following case:

        .. code-block:: python

            repo.filter_by(color="red").count()

        """
        raise NotImplementedError("Count using query not implemented.")

    def format_query(self, query: dict) -> dict:
        """Format the query to a format suitable by the repository

        You may override this method. By default,
        the query is as dictionary.

        Parameters
        ----------
        query : dict
            Query to reformat

        Examples
        --------

        Used in the following case:
        
        .. code-block:: python

            repo.filter_by(color="red")

        """
        raise NotImplementedError("Query formatting not implemented.")




from typing import Optional

from redbird.repos import CSVFileRepo
from pydantic import BaseModel

class Item(BaseModel):
    id: str
    name: str
    age: Optional[int]

def test_filecontent(tmpdir):
    file = tmpdir / "test.csv"
    repo = CSVFileRepo(filename=file, model=Item)

    repo.add(Item(id="a", name="Jack", age=20))
    repo.add(Item(id="b", name="John"))
    repo.add(Item(id="c", name="James", age=30))

    assert "id,name,age\na,Jack,20\nb,John,\nc,James,30\n" == file.read_text(encoding="UTF-8")

    repo.filter_by(id="b").update(age=40)
    assert "id,name,age\na,Jack,20\nb,John,40\nc,James,30\n" == file.read_text(encoding="UTF-8")

    repo.filter_by(id="b").delete()
    assert "id,name,age\na,Jack,20\nc,James,30\n" == file.read_text(encoding="UTF-8")

    repo.filter_by().delete()
    assert "id,name,age\n" == file.read_text(encoding="UTF-8")

def test_none(tmpdir):
    file = tmpdir / "test.csv"
    repo = CSVFileRepo(filename=file, model=Item)
    repo.add(Item(id="b", name="John"))
    assert [Item(id="b", name="John", age=None)] == repo.filter_by().all()

def test_kwds_csv(tmpdir):
    file = tmpdir / "test.csv"
    repo = CSVFileRepo(filename=file, model=Item, kwds_csv={'delimiter': '|'})

    repo.add(Item(id="a", name="Jack", age=20))
    repo.add(Item(id="b", name="John"))

    assert "id|name|age\na|Jack|20\nb|John|\n" == file.read_text(encoding="UTF-8")

    assert repo.filter_by().all() == [Item(id="a", name="Jack", age=20), Item(id="b", name="John")]

def test_dict_operations(tmpdir):
    file = tmpdir / "test.csv"
    repo = CSVFileRepo(filename=file, fieldnames=['id', 'name', 'age'])

    # Insert
    repo.add(dict(id="a", name="Jack", age=20))
    repo.add(dict(id="b", name="Jack"))
    repo.add(dict(id="c", name="James", age=30))

    # Read
    assert repo.filter_by().all() == [
        dict(id="a", name="Jack", age="20"),
        dict(id="b", name="Jack", age=None),
        dict(id="c", name="James", age="30")
    ]
    assert repo.filter_by(name="Jack").all() == [
        dict(id="a", name="Jack", age="20"),
        dict(id="b", name="Jack", age=None)
    ]
    assert repo.filter_by(age="30").all() == [
        dict(id="c", name="James", age="30"),
    ]
    assert repo.filter_by(age=30).all() == []

    # Update
    repo.filter_by(name="Jack").update(age=50)
    assert repo.filter_by().all() == [
        dict(id="a", name="Jack", age="50"),
        dict(id="b", name="Jack", age="50"),
        dict(id="c", name="James", age="30")
    ]

    # Delete
    repo.filter_by(name="Jack").delete()
    assert repo.filter_by().all() == [
        #dict(id="a", name="Jack", age="50"),
        #dict(id="b", name="Jack", age="50"),
        dict(id="c", name="James", age="30")
    ]



from operator import getitem
from typing import Any, Dict, List

from pydantic import BaseModel, Field, PrivateAttr
from redbird import BaseRepo, BaseResult
from redbird.templates import TemplateRepo
from redbird.exc import KeyFoundError
from redbird.oper import Operation
from redbird.dummy import DummySession

from redbird.utils.query import QueryMatcher

class MemoryRepo(TemplateRepo):
    """Memory repository

    This is a repository which operates in memory.
    Useful for unit tests and prototyping.

    Parameters
    ----------
    collection : list
        The collection
    model : Type
        Class of an item in the repository.
        Commonly dict or subclass of Pydantic
        BaseModel. By default dict
    id_field : str, optional
        Attribute or key that identifies each item
        in the repository.
    field_access : {'attr', 'key'}, optional
        How to access a field in an item. Either
        by attribute ('attr') or key ('item').
        By default guessed from the model.
    query : Type, optional
        Query model of the repository.
    errors_query : {'raise', 'warn', 'discard'}
        Whether to raise an exception, warn or discard
        the item in case of validation error in 
        converting data to the item model from
        the repository. By default raise 
    
    Examples
    --------
    .. code-block:: python

        repo = MemoryRepo()

    .. code-block:: python

        repo = MemoryRepo(collection=[
            {"car_type": "van", "color": "red"},
            {"car_type": "truck", "color": "red"}
        ])
    """
    #cls_result = MemoryResult
    collection: List[Any] = []

    ordered: bool = Field(default=False, const=True)
    _session = PrivateAttr()
    
    def insert(self, item):
        if self.id_field is not None:
            id_ = self.get_field_value(item, self.id_field)
            if id_ in [self.get_field_value(col_item, self.id_field) for col_item in self.collection]:
                raise KeyFoundError(f"Item {id_} already in the collection.")
        data = self.item_to_data(item)
        self.collection.append(data)

    def item_to_data(self, item):
        return item

    def query_data(self, query):
        query = QueryMatcher(query, value_getter=self.get_field_value)
        col = self.collection
        for data in col:
            if data in query:
                yield data

    def query_update(self, query:dict, values:dict):
        query = QueryMatcher(query, value_getter=self.get_field_value)
        col = self.collection
        for item in col:
            if item in query:
                for key, val in values.items():
                    self.set_field_value(item, key, val)

    def query_delete(self, query:dict):
        query = QueryMatcher(query, value_getter=self.get_field_value)
        cont = self.collection
        self.collection = [
            item 
            for item in cont 
            if not item in query
        ]

    @property
    def session(self):
        if not hasattr(self, "_session"):
            self._session = DummySession()
        return self._session



from operator import getitem
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union, TextIO

import csv, json

from pydantic import BaseModel, PrivateAttr
from redbird import BaseRepo, BaseResult
from redbird.base import Data, Item
from redbird.templates import TemplateRepo
from redbird.exc import KeyFoundError
from redbird.oper import Operation
from redbird.dummy import DummySession
from redbird.utils.query import delete_items, read_items, update_items

from redbird.utils.query import QueryMatcher

class JSONDirectoryRepo(TemplateRepo):
    """JSON directory repository

    This repository represents a directory which 
    contains JSON files. Each item represents a
    file and the ``id_field`` is the stem of the
    file.

    Parameters
    ----------
    path : Path-like
        Path to the repository directory
    model : Type
        Class of an item in the repository.
        Commonly dict or subclass of Pydantic
        BaseModel. By default dict
    id_field : str
        Attribute or key that identifies each item
        in the repository.
    field_access : {'attr', 'key'}, optional
        How to access a field in an item. Either
        by attribute ('attr') or key ('item').
        By default guessed from the model.
    query : Type, optional
        Query model of the repository.
    errors_query : {'raise', 'warn', 'discard'}
        Whether to raise an exception, warn or discard
        the item in case of validation error in 
        converting data to the item model from
        the repository. By default raise 
    kwds_json_load : dict
        Keyword arguments passed to ``json.load``. 
    kwds_json_dump : dict
        Keyword arguments passed to ``json.dump``.
        Useful for prettifying the files

    
    Examples
    --------
    .. code-block:: python

        repo = JSONDirectoryRepo(path="path/to/repo", id_field="id")

    """

    path: Path
    field_names: Optional[List[str]] = None
    kwds_json_load: dict = {}
    kwds_json_dump: dict = {}
    id_field: str

    _session = PrivateAttr()

    def insert(self, item):
        id = self.get_field_value(item, self.id_field)
        file = self.get_file_path(id)
        if file.exists():
            raise KeyFoundError("Item already exists")
        self.write_file(item)

    def query_items(self, query: dict) -> Item:
        filename_query = query.get(self.id_field, None)
        yield from read_items(self, self.read_data(filename_query), query)

    def query_update(self, query:dict, values:dict):
        filename_query = query.get(self.id_field, None)
        new_items = update_items(self, self.read_data(filename_query), query, values, return_="updated")
        self.write_files(new_items)

    def query_delete(self, query:dict):
        filename_query = query.get(self.id_field, None)
        new_items = delete_items(self, self.read_data(filename_query), query, return_="deleted")
        self.delete_files(new_items)

    def read_data(self, id_value: Union[str, Operation]) -> Iterator[Data]:
        optimized = id_value is not None
        if optimized:
            # We don't need to open all files
            if isinstance(id_value, Operation):
                raise NotImplementedError("Operation ID search not yet implemented")
            else:
                try:
                    yield self.read_file(id_value)
                except FileNotFoundError as exc:
                    raise KeyError('Item not found') from exc
        else:
            # We need to open all files
            for file in self.path.glob("*.json"):
                id = file.stem
                yield self.read_file(id)

    def read_file(self, id) -> Data:
        "Read repository file"
        filename = self.get_file_path(id)
        with open(filename, "r") as file:
            return json.load(file, **self.kwds_json_load)

    def write_files(self, items: Iterator[Item]):
        "Write the files with new content of items"
        for item in items:
            self.write_file(item)

    def write_file(self, item: Item):
        "Write to a file with new content of an item"
        id_value = self.get_field_value(item, self.id_field)
        filename = self.get_file_path(id_value)
        with open(filename, "w") as file:
            data = self.item_to_dict(item)
            json.dump(data, file, **self.kwds_json_dump)

    def delete_files(self, items: Iterator[Item]):
        for item in items:
            id_value = self.get_field_value(item, self.id_field)
            filename = self.get_file_path(id_value)
            filename.unlink()

    def get_writer(self, buff: TextIO):
        return csv.DictWriter(buff, fieldnames=self.get_headers(), **self.kwds_csv)

    def get_reader(self, buff: TextIO):
        return csv.DictReader(buff, fieldnames=self.get_headers(), **self.kwds_csv)

    def create(self, if_exists="raise"):
        "Create the repository file"
        if if_exists == "raise" and self.filename.exists():
            raise FileExistsError(f"Repository file {self.filename} already exists.")
        with open(self.filename, "w", newline="") as file:
            writer = self.get_writer(file)
            writer.writeheader()

    def data_to_item(self, data: dict) -> Item:
        # None values are converted to empty strings
        # Therefore we interpret empty strings as None
        data = {key: None if val == "" else val for key, val in data.items()}
        return super().data_to_item(data)

    def get_file_path(self, id):
        return self.path / f"{id}.json"

    def create(self):
        "Create the repository directory"
        self.path.mkdir(parents=True, exist_ok=True)

    @property
    def session(self):
        if not hasattr(self, "_session"):
            self._session = DummySession()
        return self._session



from typing import Optional

from redbird.repos import MongoRepo

import pytest
from pydantic import BaseModel

class Item(BaseModel):
    id: str
    name: str
    age: Optional[int]


class ItemWithCol(BaseModel):
    __colname__ = 'items'
    id: str
    name: str
    age: Optional[int]

def test_creation_defaults():
    pytest.importorskip("pymongo")
    repo = MongoRepo(model=ItemWithCol, uri="mongodb://localhost:27017/pytest?authSource=admin", id_field="id")

    col = repo.get_collection()
    assert col.name == "items"
    assert col.database.name == "pytest"

def test_from_client():
    pytest.importorskip("pymongo")
    from pymongo import MongoClient
    client = MongoClient("mongodb://localhost:27017/pytest?authSource=admin")
    repo = MongoRepo.from_client(model=ItemWithCol, client=client, id_field="id", database="pytest")

    col = repo.get_collection()
    assert col.name == "items"
    assert col.database.name == "pytest"

def test_creation_passed(mongo_uri):
    pytest.importorskip("pymongo")
    repo = MongoRepo(model=Item, uri=mongo_uri, database="pytest2", collection="items", id_field="id")

    col = repo.get_collection()
    assert col.name == "items"
    assert col.database.name == "pytest2"

def test_init_uri(mongo_uri):
    pytest.importorskip("pymongo")
    repo = MongoRepo(uri=mongo_uri, database="pytest", collection="items")

    col = repo.get_collection()
    assert col.name == "items"
    assert col.database.name == "pytest"

def test_init_client():
    pytest.importorskip("pymongo")
    from pymongo import MongoClient
    client = MongoClient("mongodb://localhost:27017/pytest?authSource=admin")
    repo = MongoRepo(client=client, database="pytest", collection="items")

    col = repo.get_collection()
    assert col.name == "items"
    assert col.database.name == "pytest"

def test_init_fail_missing_connection():
    pytest.importorskip("pymongo")
    from pymongo import MongoClient
    with pytest.raises(TypeError):
        repo = MongoRepo(database="pytest", collection="items")


# Test deprecated
def test_deprecated():
    pytest.importorskip("pymongo")
    from pymongo import MongoClient
    client = MongoClient("mongodb://localhost:27017/pytest?authSource=admin")

    repo_from_client = MongoRepo.from_uri(uri="mongodb://localhost:27017/pytest?authSource=admin", database="pytest", collection="items")
    repo_from_uri = MongoRepo.from_client(client=client, database="pytest", collection="items")

    col = repo_from_client.get_collection()
    assert col.name == "items"
    assert col.database.name == "pytest"

    col = repo_from_uri.get_collection()
    assert col.name == "items"
    assert col.database.name == "pytest"


import warnings

class KeyFoundError(Exception):
    """Typically raised in insertion.
    
    Raised when an item for given key/id field
    is found when it was not expected.
    
    """

class ItemToDataError(ValueError):
    "Raise when converting item to data failed"

class DataToItemError(ValueError):
    "Raise when converting data to item failed"


class ConversionWarning(UserWarning):
    "Converting data to item or item to data failed non-fatally"


def _handle_conversion_error(repo, data):
    errors_query = repo.errors_query
    if errors_query == "raise":
        raise
    elif errors_query == "warn":
        warnings.warn(f'Converting data to item failed: \n{data}', ConversionWarning)




from abc import ABC
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from redbird.base import BaseResult

class Operation(ABC):
    """Field operation
    """
    __py_magic__: str
    __formatter__ : str
    def __init__(self, value):
        self.value = value

    def _get_formatter(self, result:'BaseResult') -> Callable:
        return getattr(result, self.__formatter__)

    def evaluate(self, value):
        return getattr(self, self.__py_magic__)(value)

class GreaterThan(Operation):
    __py_magic__ = "__gt__"
    __formatter__ = "format_greater_than"
    def __gt__(self, value):
        return value > self.value

class LessThan(Operation):
    __py_magic__ = "__lt__"
    __formatter__ = "format_less_than"
    def __lt__(self, value):
        return value < self.value

class GreaterEqual(Operation):
    __py_magic__ = "__ge__"
    __formatter__ = "format_greater_equal"
    def __ge__(self, value):
        return value >= self.value

class LessEqual(Operation):
    __py_magic__ = "__le__"
    __formatter__ = "format_less_equal"
    def __le__(self, value):
        return value <= self.value

class NotEqual(Operation):
    __py_magic__ = "__ne__"
    __formatter__ = "format_not_equal"
    def __ne__(self, value):
        return value != self.value

class Between(Operation):

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def evaluate(self, value):
        return self.start <= value <= self.end

class In(Operation):
    __py_magic__ = "__contains__"

    def evaluate(self, value):
        return value in self.value

class _Skip(Operation):
    """Field operator that does not affect filtering.
    This is just a convenient placeholder."""

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            # Create the instance
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        pass

    def evaluate(self, value):
        return True

def greater_than(value):
    return GreaterThan(value)
    
def less_than(value):
    return LessThan(value)

def greater_equal(value):
    return GreaterEqual(value)
    
def less_equal(value):
    return LessEqual(value)

def not_equal(value):
    return NotEqual(value)

def between(start, end, none_as_open=False):
    if none_as_open:
        # Handle Nones
        if start is None and end is None:
            return skip
        elif end is None:
            return greater_equal(start)
        elif start is None:
            return less_equal(end)
    return Between(start, end)

def in_(values):
    return In(values)

skip = _Skip()



import pytest
from redbird import BaseRepo, BaseResult

def test_result_incomplete():
    with pytest.raises(TypeError) as excinfo:
        class MyResults(BaseResult):
            ...
        MyResults()
    assert str(excinfo.value.args[0]) == "Can't instantiate abstract class MyResults with abstract methods delete, query_data, update"

def test_repo_incomplete():
    with pytest.raises(TypeError) as excinfo:
        class MyRepo(BaseRepo):
            ...
        MyRepo()
    assert str(excinfo.value.args[0]) in (
        # Python <3.10
        "Can't instantiate abstract class MyRepo with abstract methods insert",
        # Python >=3.10
        "Can't instantiate abstract class MyRepo with abstract method insert",
    )




class DummySession:
    """Dummy session

    Imitates similar session objects as SQLAlchemy's
    session in order to avoid code changes if
    in-memory repository is used.
    """

    def close(self):
        "Close the connection(s)/client(s)/engine(s)"
        ...
    
    def remove(self):
        "Close the connection(s)/client(s)/engine(s) so that they can be recreated"
        ...

    def get_bind(self, bind_key=None):
        "Get connection/client/engine for given key"
        ...




from typing import TYPE_CHECKING, Any, Dict, Optional, Union


from pydantic import BaseModel, Field, ValidationError

from redbird.base import BaseResult, BaseRepo
from redbird.exc import KeyFoundError
from redbird.oper import Between, GreaterEqual, GreaterThan, In, LessEqual, LessThan, NotEqual, Operation, skip
from redbird.templates import TemplateRepo

if TYPE_CHECKING:
    from pymongo import MongoClient
    from pymongo.database import Database
    from pymongo.collection import Collection    

class MongoSession:

    """MongoDB session

    Similar as SQLAlchemy Session but works slightly 
    different due to technical details

    MongoClient is analogous to sqlalchemy.engine.Engine but 
    MongoClient is not fork safe: https://pymongo.readthedocs.io/en/stable/faq.html#is-pymongo-fork-safe
    SQLAlchemy Engine is fork safe if no connections are made.

    This object is fork safe if no clients are open (call remove to close all of them).
    """

    url: str
    binds: Dict[str, str]

    _bind: 'MongoClient'
    _binds: Dict[str, 'MongoClient']

    def __init__(self, url=None, binds:Dict[str, str]=None, client=None):
        self.url = url
        self.binds = binds if binds is not None else {}

        self._binds = {}
        self._bind = client

    @property
    def client(self):
        if self._bind is None:
            self._bind = self.create_client()
        return self._bind

    def create_client(self, url=None):
        import pymongo
        url = self.url if url is None else url
        return pymongo.MongoClient(url)

    def close(self):
        "Close client and all binds"
        if self._bind is not None:
            self._bind.close()
        for client in self._binds.values():
            client.close()

    def remove(self):
        """Remove the connection

        Similar in some ways to SQLAlchemy's Session.remove.
        This method will close the clients and remove them.
        Next time the client is accessed, a new client is 
        created. Meant to be safe for multi-threaded apps.
        """
        self.close()

        self._bind = None
        self._binds = {}

    def get_bind(self, bind_key=None) -> 'MongoClient':
        "Get client associated with the mapper"
        if bind_key is None:
            # return default connection
            return self.client
        else:
            bind_key = getattr(bind_key, "__bind_key__", bind_key)
            url = self.binds.get(bind_key, self.url)
            if url not in self._binds:
                self._binds[url] = self.create_client(url)
            return self._binds[url]

class MongoRepo(TemplateRepo):
    """MongoDB Repository

    Parameters
    ----------
    uri : str, optional
        Connection string to the database.
        Pass uri, client or session.
    client : mongodb.MongoClient, optional
        MongoDB client for the connection.
        Pass uri, client or session.
    model : Type
        Class of an item in the repository.
        Commonly dict or subclass of Pydantic
        BaseModel. By default dict
    id_field : str, optional
        Attribute or key that identifies each item
        in the repository.
    field_access : {'attr', 'key'}, optional
        How to access a field in an item. Either
        by attribute ('attr') or key ('item').
        By default guessed from the model.
    query : Type, optional
        Query model of the repository.
    errors_query : {'raise', 'warn', 'discard'}
        Whether to raise an exception, warn or discard
        the item in case of validation error in 
        converting data to the item model from
        the repository. By default raise 
    session : Session, Any
        A MongoDB session object that should
        have at least ``client`` attribute.
        Pass uri, client or session.

    Examples
    --------
    .. code-block:: python

        repo = MongoRepo(uri="mongodb://localhost:27017/mydb?authSource=admin", collection="mycol")

    .. code-block:: python

        repo = MongoRepo(uri="mongodb://localhost:27017", database="mydb", collection="mycol")

    .. code-block:: python

        from pymongo import MongoClient
        repo = MongoRepo(client=MongoClient("mongodb://localhost:27017"))
    """
    # cls_result = MongoResult
    default_id_field = "_id"
    cls_session = MongoSession

    __operators__ = {
        GreaterThan: "$gt",
        LessThan: "$lt",
        GreaterEqual: "$gte",
        LessEqual: "$lte",
        NotEqual: "$ne",
        In: "$in",
    }
    session: Any
    database: Optional[str]
    collection: Optional[str]

    ordered: bool = Field(default=True, const=True)

    def __init__(self, *args, uri=None, client=None, session=None, **kwargs):
        if uri is not None:
            session = MongoSession(url=uri)
        if client is not None:
            session = MongoSession(client=client)
        if session is None:
            raise TypeError(
                "Cannot determine the connection. "
                "Consider passing 'uri', 'client' or 'session'."
            )
        super().__init__(*args, session=session, **kwargs)

    @classmethod
    def from_uri(cls, *args, uri, **kwargs):
        kwargs["session"] = MongoSession(url=uri)
        return cls(*args, **kwargs)

    @classmethod
    def from_client(cls, *args, client, **kwargs):
        kwargs["session"] = MongoSession(client=client)
        return cls(*args, **kwargs)

    def insert(self, item):
        from pymongo.errors import DuplicateKeyError
        col = self.get_collection()
        doc = self.item_to_data(item)
        try:
            col.insert_one(doc)
        except DuplicateKeyError as exc:
            raise KeyFoundError(f"Document {self.get_field_value(item, self.id_field)} already exists.") from exc

    def upsert(self, item):
        col = self.get_collection()
        doc = self.item_to_data(item)
        col.update_one({"_id": self.get_field_value(item, self.id_field)}, {"$set": doc}, upsert=True)

    def filter_by(self, **kwargs) -> BaseResult:
        # Rename id_field --> "_id"
        if self.id_field in kwargs:
            kwargs["_id"] = kwargs.pop(self.id_field)
        return super().filter_by(**kwargs)

    def get_collection(self) -> 'Collection':
        "Get the MongoDB collection object"
        col_name = self.model.__colname__ if hasattr(self.model, "__colname__") else self.collection
        database = self.get_database()
        return database[col_name]

    def get_database(self) -> 'Database':
        "Get the MongoDB database object"
        client = self.get_client()
        return client[self.database] if self.database is not None else client.get_default_database()

    def get_client(self) -> 'MongoClient':
        "Get the MongoDB client"
        #! TODO: Support __bind_key__
        return self.session.get_bind(self.model)

    def _format_dict(self, item:dict) -> BaseModel:
        try:
            return self.model(**item)
        except ValidationError as exc:
            raise ValidationError(f"Formatting for {item[self.id_field]} failed.") from exc

    def data_to_item(self, json:dict):
        # Rename _id to whatever is as id_field
        json[self.id_field] = json.pop("_id")
        return self.model(**json)

    def item_to_data(self, item:BaseModel):
        json = self.item_to_dict(item)
        # Rename whatever is as id_field to _id
        json["_id"] = json.pop(self.id_field)
        return json

# Query based
    def query_data(self, query):
        col = self.get_collection()
        for data in col.find(query):
            yield data

    def query_data_limit(self, query, n:int):
        "Get n first items"
        col = self.get_collection()
        return [
            self.data_to_item(item)
            for item in col.find(query).limit(n)
        ]

    def query_update(self, query, values):
        col = self.get_collection()
        col.update_many(query, {"$set": values})

    def query_delete(self, query):
        col = self.get_collection()
        col.delete_many(query)

    def query_count(self, query):
        col = self.get_collection()
        return col.count_documents(query)

    def format_query(self, query):
        return {
            key: self._get_query_value(val)
            for key, val in query.items()
            if val is not skip
        }

    def _get_query_value(self, value):
        if isinstance(value, Operation):
            type_ = type(value)
            if type_ in self.__operators__:
                return {self.__operators__[type_]: value.value}
            elif type_ == Between:
                return {
                    '$gte': value.start,
                    '$lte': value.end,
                }
            else:
                raise NotImplementedError("MongoRepo does not yet support operator: {type_}")
        else:
            return value




import configparser
from typing import Optional
import re, os
import json
from dotenv import load_dotenv

import pytest
import responses
import requests
import mongomock
from redbird.repos.csv import CSVFileRepo
from redbird.repos.json import JSONDirectoryRepo
from redbird.repos.rest import RESTRepo
from redbird.repos.sqlalchemy import SQLRepo
from redbird.repos.memory import MemoryRepo
from redbird.repos.mongo import MongoRepo
from redbird.oper import greater_equal, greater_than, less_equal, less_than, not_equal

from pydantic import BaseModel, Field

# ------------------------
# TEST ITEMS
# ------------------------

class PydanticItem(BaseModel):
    __colname__ = 'items'
    id: str
    name: str
    age: Optional[int]

class PydanticItemORM(BaseModel):
    id: str
    name: str
    age: Optional[int]
    class Config:
        orm_mode = True

class MongoItem(BaseModel):
    __colname__ = 'items'
    id: str  = Field(alias="_id")
    name: str
    age: int

try:
    from sqlalchemy import Column, String, Integer, create_engine
    from sqlalchemy.orm import declarative_base
except ImportError:
    # Has no SQLAlchemy
    pass
else:
    SQLBase = declarative_base()

    class SQLItem(SQLBase):
        __tablename__ = 'items'
        id = Column(String, primary_key=True)
        name = Column(String)
        age = Column(Integer)

        def __eq__(self, other):
            if not isinstance(other, SQLItem):
                return False
            return other.id == self.id and other.name == self.name and other.age == self.age

# ------------------------
# MOCK
# ------------------------

def get_mongo_uri():
    load_dotenv()
    pytest.importorskip("pymongo")
    if "MONGO_CONN" not in os.environ:
        pytest.skip()
    return os.environ["MONGO_CONN"]

class RESTMock:

    def __init__(self):
        self.repo = MemoryRepo(model=PydanticItem, id_field="id")
    
    def post(self, request):
        data = json.loads(request.body)
        self.repo.add(data)
        return (200, {}, b"")

    def patch(self, request):
        data = json.loads(request.body)
        params = self.get_params(request)
        self.repo.filter_by(**params).update(**data)
        return (200, {}, b"")

    def patch_one(self, request):
        id = self.get_id(request)
        data = json.loads(request.body)
        assert "id" not in data
        self.repo.get_by(id).update(**data)
        return (200, {}, b"")

    def put(self, request):
        data = json.loads(request.body)
        item = self.repo.model(**data)
        self.repo.replace(item)
        return (200, {}, b"")

    def delete(self, request):
        params = self.get_params(request)
        self.repo.filter_by(**params).delete()
        return (200, {}, b"")

    def delete_one(self, request):
        id = self.get_id(request)
        del self.repo[id]
        return (200, {}, b"")

    def get(self, request):
        params = self.get_params(request)
        data = self.repo.filter_by(**params).all()
        data = [item.dict() for item in data]
        return (200, {"Content-Type": "application/json"}, json.dumps(data))

    def get_one(self, request):
        id = self.get_id(request)
        data = self.repo[id].dict()
        return (200, {"Content-Type": "application/json"}, json.dumps(data))

    def get_params(self, req):
        return {
            key: int(val) if val.isdigit() else val 
            for key, val in req.params.items()
        }

    def get_id(self, req):
        parts = req.url.rsplit("api/items/", 1)
        return parts[-1] if len(parts) > 1 else None

    def add_routes(self, rsps):
        rsps.add_callback(
            responses.POST, 
            'http://localhost:5000/api/items',
            callback=self.post,
            content_type='application/json',
        )
        rsps.add_callback(
            responses.PATCH, 
            re.compile('http://localhost:5000/api/items/[a-zA-Z]+'),
            callback=self.patch_one,
        )
        rsps.add_callback(
            responses.PATCH, 
            re.compile('http://localhost:5000/api/items?[a-zA-Z=_]+'),
            callback=self.patch,
        )

        rsps.add_callback(
            responses.PUT, 
            re.compile('http://localhost:5000/api/items'),
            callback=self.put,
        )

        rsps.add_callback(
            responses.DELETE, 
            'http://localhost:5000/api/items',
            callback=self.delete,
        )
        rsps.add_callback(
            responses.DELETE, 
            re.compile('http://localhost:5000/api/items/[a-zA-Z]+'),
            callback=self.delete_one,
        )

        rsps.add_callback(
            responses.GET, 
            re.compile('http://localhost:5000/api/items/[a-zA-Z]+'),
            callback=self.get_one,
        )
        rsps.add_callback(
            responses.GET, 
            re.compile('http://localhost:5000/api/items'),
            callback=self.get,
        )


def get_repo(type_, tmpdir):
    if type_.startswith("sql-"):
        pytest.importorskip("sqlalchemy")
        from sqlalchemy import Column, String, Integer, create_engine
    if type_ == "memory":
        repo = MemoryRepo(model=PydanticItem, id_field="id")

    elif type_ == "memory-dict":
        repo = MemoryRepo(model=dict, id_field="id")

    elif type_ == "csv":
        repo = CSVFileRepo(model=PydanticItem, id_field="id", filename=str(tmpdir / "repo.csv"))
        repo.create()

    elif type_ == "json-dir":
        repo = JSONDirectoryRepo(model=PydanticItem, id_field="id", path=str(tmpdir))

    elif type_ == "sql-dict":
        engine = create_engine('sqlite://')
        engine.execute("""CREATE TABLE pytest (
            id TEXT PRIMARY KEY,
            name TEXT,
            age INTEGER
        )""")
        repo = SQLRepo(engine=engine, table="pytest", id_field="id")

    elif type_ == "sql-pydantic":
        engine = create_engine('sqlite://')
        engine.execute("""CREATE TABLE pytest (
            id TEXT PRIMARY KEY,
            name TEXT,
            age INTEGER
        )""")
        repo = SQLRepo(model=PydanticItem, engine=engine, table="pytest", id_field="id")
        #SQLItem.__table__.create(bind=repo.session.bind)

    elif type_ == "sql-orm":
        engine = create_engine('sqlite://')
        repo = SQLRepo(model_orm=SQLItem, engine=engine, table="items", id_field="id")
        repo.create()

    elif type_ == "sql-pydantic-orm":
        engine = create_engine('sqlite://')
        repo = SQLRepo(model=PydanticItemORM, model_orm=SQLItem, engine=engine, id_field="id")
        SQLItem.__table__.create(bind=repo.session.bind)

    elif type_ == "mongo-mock":
        pytest.importorskip("pymongo")
        repo = MongoRepo(model=PydanticItem, uri="mongodb://localhost:27017/pytest?authSource=admin", database="pytest", collection="items", id_field="id")
 
    elif type_ == "mongo":
        pytest.importorskip("pymongo")
        repo = MongoRepo(model=PydanticItem, uri=get_mongo_uri(), database="pytest", collection="items", id_field="id")

        # Empty the collection
        import pymongo

        client = pymongo.MongoClient(repo.session.url)
        col_name = repo.model.__colname__
        db = client.get_default_database()
        col = db[col_name]
        col.delete_many({})

    elif type_ == "http-rest":
        repo = RESTRepo(model=PydanticItem, url="http://localhost:5000/api/items", id_field="id")

    return repo

# ------------------------
# FIXTURES
# ------------------------

@pytest.fixture
def repo(request, tmpdir):
    repo = get_repo(request.param, tmpdir)
    if request.param == "http-rest":
        api = RESTMock()
        with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
            api.add_routes(rsps)
            yield repo
    elif request.param == "mongo-mock":
        with mongomock.patch(servers=(('localhost', 27017),)):
            yield repo
    else:
        yield repo



from redbird.oper import skip, Between, LessEqual, between, GreaterEqual


def test_between_open_left():
    oper = between(None, 30, none_as_open=True)
    assert isinstance(oper, LessEqual)
    assert oper.value == 30

def test_between_open_right():
    oper = between(30, None, none_as_open=True)
    assert isinstance(oper, GreaterEqual)
    assert oper.value == 30

def test_between_open_both():
    oper = between(None, None, none_as_open=True)
    assert oper is skip

def test_between_strict_none():
    oper = between(None, None)
    assert isinstance(oper, Between)
    assert oper.start is None
    assert oper.end is None



import pytest
from redbird.repos import SQLRepo
from pydantic import BaseModel

class MyItem(BaseModel):
    id: str
    name: str
    age: int

class MyItemWithORM(BaseModel):
    id: str
    name: str
    age: int
    class Config:
        orm_mode = True

try:
    from sqlalchemy.orm import declarative_base
    from sqlalchemy import Column, String, Integer
except ImportError:
    pass
else:
    SQLBase = declarative_base()

    class SQLItem(SQLBase):
        __tablename__ = 'pytest'
        id = Column(String, primary_key=True)
        name = Column(String)
        age = Column(Integer)

        def __eq__(self, other):
            if not isinstance(other, SQLItem):
                return False
            return other.id == self.id and other.name == self.name and other.age == self.age

def test_init_engine():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    engine.execute("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""")
    repo = SQLRepo(engine=engine, table="pytest")
    assert repo.model_orm.__table__.name == "pytest"
    assert all(hasattr(repo.model_orm, col) for col in ("id", "name", "age"))

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [dict(id="a", name="Jack", age=500)]

def test_init_conn_string(tmpdir):
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    with tmpdir.as_cwd() as old_dir:
        conn_string = 'sqlite:///pytest.db'
        engine = create_engine(conn_string)
        engine.execute("""CREATE TABLE pytest (
            id TEXT PRIMARY KEY,
            name TEXT,
            age INTEGER
        )""")
        repo = SQLRepo(conn_string=conn_string, table="pytest")
        assert repo.model_orm.__table__.name == "pytest"
        assert all(hasattr(repo.model_orm, col) for col in ("id", "name", "age"))

        repo.add({"id": "a", "name": "Jack", "age": 500})
        assert list(repo) == [dict(id="a", name="Jack", age=500)]

def test_init_session():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    from sqlalchemy.orm import Session
    engine = create_engine('sqlite://')
    session = Session(engine)
    engine.execute("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""")
    repo = SQLRepo(session=session, table="pytest")
    assert repo.model_orm.__table__.name == "pytest"
    assert all(hasattr(repo.model_orm, col) for col in ("id", "name", "age"))

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [dict(id="a", name="Jack", age=500)]

def test_init_model():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    engine.execute("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""")
    repo = SQLRepo(model=MyItem, engine=engine, table="pytest")
    assert repo.model is MyItem

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [MyItem(id="a", name="Jack", age=500)]

def test_init_model_orm_mode():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    engine.execute("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""")
    repo = SQLRepo(model=MyItemWithORM, engine=engine, table="pytest")
    assert repo.model is MyItemWithORM

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [MyItemWithORM(id="a", name="Jack", age=500)]

def test_init_orm():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    engine.execute("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""")
    repo = SQLRepo(model_orm=SQLItem, reflect_model=True, engine=engine)
    assert repo.model_orm is SQLItem
    assert issubclass(repo.model, BaseModel)

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [repo.model(id="a", name="Jack", age=500)]

def test_init_model_and_orm():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    engine.execute("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""")
    repo = SQLRepo(model=MyItem, model_orm=SQLItem, engine=engine)
    assert repo.model_orm is SQLItem
    assert issubclass(repo.model, BaseModel)

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [MyItem(id="a", name="Jack", age=500)]

def test_init_dict():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    engine.execute("""CREATE TABLE pytest (
        id TEXT PRIMARY KEY,
        name TEXT,
        age INTEGER
    )""")

    repo = SQLRepo(model=dict, engine=engine, table="pytest")
    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [{"id": "a", "name": "Jack", "age": 500}]

def test_init_fail_missing_connection():
    pytest.importorskip("sqlalchemy")
    with pytest.raises(TypeError):
        SQLRepo(model=MyItem, table="my_table")

def test_init_deprecated(tmpdir):
    with tmpdir.as_cwd() as old_dir:
        pytest.importorskip("sqlalchemy")
        from sqlalchemy import create_engine
        engine = create_engine('sqlite:///pytest.db')
        engine.execute("""CREATE TABLE pytest (
            id TEXT PRIMARY KEY,
            name TEXT,
            age INTEGER
        )""")

        repo = SQLRepo.from_engine(model=dict, engine=engine, table="pytest")
        repo.add({"id": "a", "name": "Jack", "age": 500})
        assert list(repo) == [{"id": "a", "name": "Jack", "age": 500}]
        
        repo = SQLRepo.from_connection_string(model=dict, conn_string="sqlite:///pytest.db", table="pytest")
        repo.add({"id": "b", "name": "Jack", "age": 5000})
        assert list(repo) == [{"id": "a", "name": "Jack", "age": 500}, {"id": "b", "name": "Jack", "age": 5000}]


def test_init_reflect_model_without_primary_key():
    # https://stackoverflow.com/a/23771348/13696660
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    engine = create_engine('sqlite://')
    with pytest.raises(KeyError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable", if_missing="create")
    
    engine.execute("""CREATE TABLE pytest (
        id TEXT,
        name TEXT,
        age INTEGER
    )""")
    with pytest.raises(KeyError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable")

def test_init_create_table():
    pytest.importorskip("sqlalchemy")
    from sqlalchemy import create_engine
    from sqlalchemy.exc import NoSuchTableError
    engine = create_engine('sqlite://')
    with pytest.raises(NoSuchTableError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable", id_field="id")
    with pytest.raises(NoSuchTableError):
        repo = SQLRepo(model=MyItem, engine=engine, table="mytable", id_field="id", if_missing="raise")

    repo = SQLRepo(model=MyItem, engine=engine, table="mytable", if_missing="create", id_field="id")
    assert issubclass(repo.model, BaseModel)

    repo.add({"id": "a", "name": "Jack", "age": 500})
    assert list(repo) == [MyItem(id="a", name="Jack", age=500)]



from abc import abstractmethod, ABC
from operator import getitem, setitem
from textwrap import dedent, indent, shorten
from typing import Any, ClassVar, Dict, Generator, Iterator, List, Mapping, Optional, Tuple, Type, TypeVar, Union
from dataclasses import dataclass
import warnings

from pydantic import BaseModel, Field

from redbird.exc import ConversionWarning, DataToItemError, KeyFoundError, ItemToDataError, _handle_conversion_error
from redbird.utils.case import to_case

from .oper import Operation

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


Item = TypeVar("Item")
Data = TypeVar("Data")

class BasicQuery(BaseModel, extra="allow"):
    ...

class DummySession:
    ...

class BaseResult(ABC):
    """Abstract filter result

    Result classes add additional alchemy
    to Red Bird providing convenient ways
    to read, modify or delete data. 
    
    Subclass of BaseRepo should also have custom
    subclass of BaseResult as cls_result attribute.
    """

    query_: dict
    repo: 'BaseRepo'

    def __init__(self, query:dict=None, repo:'BaseRepo'=None):
        self.repo = repo
        self.query_ = self.format_query(query)
        
    def first(self) -> Item:
        "Return first item"
        for item in self.query():
            return item
    
    def last(self) -> Item:
        "Return last item"
        item = None
        for item in self.query():
            pass
        return item

    def all(self) -> List[Item]:
        "Return all items"
        return list(self.query())

    def limit(self, n:int) -> List[Item]:
        "Return n items"
        items = []
        for i, item in enumerate(self.query(), start=1):
            if i > n:
                break
            items.append(item)
        return items
        
    def query(self) -> Iterator[Item]:
        "Get actual result"
        for data in self.query_data():
            try:
                yield self.repo.data_to_item(data)
            except ValueError:
                _handle_conversion_error(self.repo, data)

    @abstractmethod
    def query_data(self) -> Iterator[Data]:
        "Get actual result"
        ...

    def __iter__(self) -> Iterator[Item]:
        return self.query()

    @abstractmethod
    def update(self, **kwargs):
        "Update the resulted items"

    @abstractmethod
    def delete(self):
        "Delete the resulted items"

    def replace(self, __values:dict=None, **kwargs):
        "Replace the existing item(s) with given"
        if __values is not None:
            kwargs.update(__values)
        if self.count() > 1:
            raise KeyError("You may only replace one item.")
        self.delete()
        self.repo.add(kwargs)

    def count(self) -> int:
        "Count the resulted items"
        return len(list(self))

    def format_query(self, query:dict) -> dict:
        "Turn the query to a form that's understandable by the underlying database"
        qry = self.repo.query_model(**query)
        return qry.format(self.repo) if hasattr(qry, "format") else qry.dict()

class BaseRepo(ABC, BaseModel):
    """Abstract Repository

    Base class for the repository pattern.

    Parameters
    ----------
    model : Type
        Class of an item in the repository.
        Commonly dict or subclass of Pydantic
        BaseModel. By default dict
    id_field : str, optional
        Attribute or key that identifies each item
        in the repository.
    field_access : {'attr', 'key'}, optional
        How to access a field in an item. Either
        by attribute ('attr') or key ('item').
        By default guessed from the model.
    query : Type, optional
        Query model of the repository.
    errors_query : {'raise', 'warn', 'discard'}
        Whether to raise an exception, warn or discard
        the item in case of validation error in 
        converting data to the item model from
        the repository. By default raise 
    """
    cls_result: ClassVar[Type[BaseResult]]

    id_field: Optional[str]
    model: Type = dict
    
    query_model: Optional[Type[BaseModel]] = BasicQuery

    errors_query: Literal['raise', 'warn', 'discard'] = 'raise'
    field_access: Literal['attr', 'key', 'infer'] = 'infer'

    # Attributes that specifies how the repo behaves
    ordered: bool = Field(default=False, const=True)

    def __iter__(self) -> Iterator[Item]:
        "Iterate over the repository"
        return iter(self.filter_by().all())

    def __getitem__(self, id) -> Item:
        "Get item from the repository using ID"
        qry = {self.id_field: id}
        item = self.filter_by(**qry).first()
        if item is None:
            raise KeyError(f"Item {id} not found.")
        return item

    def __delitem__(self, id):
        "Delete item from the repository using ID"
        qry = {self.id_field: id}
        rows = self.filter_by(**qry).delete()
        if isinstance(rows, int):
            # Returned value is number of rows deleted
            if rows == 0:
                raise KeyError(f"Index {id} not found.")

    def __setitem__(self, id, attrs:dict):
        "Update given item"
        qry = {self.id_field: id}
        self.filter_by(**qry).update(**attrs)

# Item based
    def add(self, item: Item, if_exists="raise"):
        "Add an item to the repository"
        item = self.to_item(item)
        if if_exists == "raise":
            self.insert(item)
        elif if_exists == "update":
            self.upsert(item)
        elif if_exists == "ignore":
            try:
                self.insert(item)
            except KeyFoundError:
                pass
        else:
            raise ValueError(f"Invalid value: {if_exists}")

    @abstractmethod
    def insert(self):
        """Insert item to the repository
        
        Parameters
        ----------
        item: instance of model
            Item to insert to the repository

        Examples
        --------
        .. code-block:: python

            repo.insert(Item(id="a", color="red"))
        """
        ...

    def upsert(self, item: Item):
        """Upsert item to the repository

        Upsert is an insert if the item
        does not exist in the repository
        and update if it does.
        
        Parameters
        ----------
        item: instance of model
            Item to upsert to the repository

        Examples
        --------
        .. code-block:: python

            repo.upsert(Item(id="a", color="red"))
        """
        try:
            self.insert(item)
        except KeyFoundError:
            self.update(item)

    def delete(self, item: Item):
        """Delete item from the repository
        
        Parameters
        ----------
        item: instance of model
            Item to delete from the repository

        Examples
        --------
        .. code-block:: python

            repo.delete(Item(id="a", color="red"))
        """
        id_ = self.get_field_value(item, self.id_field)
        del self[id_]

    def update(self, item: Item):
        """Update item in the repository
        
        Parameters
        ----------
        item: instance of model
            Item to update in the repository

        Examples
        --------
        .. code-block:: python

            repo.update(Item(id="a", color="red"))
        """
        qry = {self.id_field: self.get_field_value(item, self.id_field)}
        values = self.item_to_dict(item)
        # We don't update the ID
        values.pop(self.id_field)
        self.filter_by(**qry).update(**values)

    def replace(self, item: Item):
        "Update an item in the repository"
        self.delete(item)
        self.add(item)

    def item_to_dict(self, item: Item) -> dict:
        if isinstance(item, dict):
            return item
        elif isinstance(item, BaseModel):
            # Pydantic model
            return item.dict(exclude_unset=True)
        else:
            return dict(**item)

# Keyword arguments
    
    def get_by(self, id):
        "Get item based on ID but returns result for further operations"
        qry = {self.id_field: id}
        return self.filter_by(**qry)

    def filter_by(self, **kwargs) -> BaseResult:
        """Filter the repository
        
        Parameters
        ----------
        **kwargs : dict
            Query which is used to conduct 
            furher operation.

        Examples
        --------
        .. code-block:: python

            repo.filter_by(color="red")
        """
        return self.cls_result(query=kwargs, repo=self)

    def data_to_item(self, data:Data) -> Item:
        "Turn object from repo (row, doc, dict, etc.) to item"
        if isinstance(data, self.model):
            # Already the right type
            return data
        if not isinstance(data, Mapping):
            # data is namespace-like
            data = vars(data)
        try:
            return self.model(**data)
        except Exception as exc:
            raise DataToItemError(f"Could not transform {data}") from exc

    def to_item(self, obj) -> Item:
        "Turn an object to item"
        if isinstance(obj, self.model):
            return obj
        elif isinstance(obj, dict):
            return self.model(**obj)
        elif isinstance(obj, (list, tuple, set)):
            return self.model(*obj)
        else:
            raise TypeError(f"Cannot cast {type(obj)} to {self.model}")

    def get_field_value(self, item: Item, key):
        """Utility method to get key's value from an item
        
        If item's fields are accessed via attribute,
        getattr is used. If fields are accessed via 
        items, getitem is used.
        """
        if self.field_access == "infer":
            field_access = "key" if hasattr(self.model, "__getitem__") else "attr"
        else:
            field_access = self.field_access
        func = {
            "attr": getattr,
            "key": getitem,
        }[field_access]
        
        return func(item, key)

    def set_field_value(self, item: Item, key, value):
        """Utility method to set field's value in an item
        
        If item's fields are accessed via attribute,
        setattr is used. If fields are accessed via 
        items, setitem is used.
        """
        if self.field_access == "infer":
            field_access = "key" if hasattr(self.model, "__getitem__") else "attr"
        else:
            field_access = self.field_access
        func = {
            "attr": setattr,
            "key": setitem,
        }[field_access]
        
        func(item, key, value)




from operator import getitem
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union, TextIO

import csv

from pydantic import BaseModel, Field, PrivateAttr
from redbird import BaseRepo, BaseResult
from redbird.base import Data, Item
from redbird.templates import TemplateRepo
from redbird.exc import KeyFoundError
from redbird.oper import Operation
from redbird.dummy import DummySession
from redbird.utils.query import delete_items, read_items, update_items

from redbird.utils.query import QueryMatcher

class CSVFileRepo(TemplateRepo):
    """CSV file repository

    This repository has a CSV (comma-separated values) 
    file as a data store. Each item represents a row
    in the file.

    Parameters
    ----------
    filename : path-like
        The repository file
    fieldnames : list of str
        Names of the columns in the CSV file.
        If unspecified, model's fields are used
        instead
    model : Type
        Class of an item in the repository.
        Commonly dict or subclass of Pydantic
        BaseModel. By default dict
    id_field : str, optional
        Attribute or key that identifies each item
        in the repository.
    field_access : {'attr', 'key'}, optional
        How to access a field in an item. Either
        by attribute ('attr') or key ('item').
        By default guessed from the model.
    query : Type, optional
        Query model of the repository.
    errors_query : {'raise', 'warn', 'discard'}
        Whether to raise an exception, warn or discard
        the item in case of validation error in 
        converting data to the item model from
        the repository. By default raise 
    kwds_csv : dict
        Keyword arguments used to create 
        ``csv.DictWriter`` and ``csv.DictReader``
    
    Examples
    --------

    .. code-block:: python

        repo = CSVFileRepo(filepath="path/to/repo", fieldnames=['id', 'name', 'age'])
    """

    filename: Path
    fieldnames: Optional[List[str]] = None
    kwds_csv: dict = {}

    _session = PrivateAttr()
    ordered: bool = Field(default=True, const=True)

    def insert(self, item):
        file_non_zero = self.filename.exists() and self.filename.stat().st_size > 0
        if not file_non_zero:
            self.create(if_exists="ignore")
        if self.id_field is not None:
            id_ = self.get_field_value(item, self.id_field)
            if id_ in [self.get_field_value(col_item, self.id_field) for col_item in self.filter_by().all()]:
                raise KeyFoundError(f"Item {id_} already in the collection.")
        self.append_file(item)

    def query_items(self, query) -> Item:
        yield from read_items(self, self.read_file(), query)

    def query_update(self, query:dict, values:dict):
        new_items = update_items(self, self.read_file(), query, values, return_='all')
        self.write_file(list(new_items))

    def query_delete(self, query:dict):
        new_items = delete_items(self, self.read_file(), query, return_='remained')
        self.write_file(list(new_items))

    @property
    def session(self):
        if not hasattr(self, "_session"):
            self._session = DummySession()
        return self._session

    def get_headers(self) -> List[str]:
        "Get headers of the CSV file (using the model)"
        if self.fieldnames is not None:
            return self.fieldnames
        elif hasattr(self.model, "__fields__"):
            return list(self.model.__fields__)
        else:
            raise TypeError("Cannot determine CSV headers")

    def read_file(self) -> Iterator[dict]:
        "Read repository file"
        with open(self.filename, "r") as file:
            reader = self.get_reader(file)

            # Skip headers
            next(reader, None)

            # Iterate rows
            for data in reader:
                yield data

    def write_file(self, items: List):
        "Write the file with new content of items"
        with open(self.filename, "w", newline="") as file:
            writer = self.get_writer(file)
            writer.writeheader()
            for item in items:
                d = self.item_to_dict(item)
                writer.writerow(d)

    def append_file(self, item: Item):
        "Write a single item at the end of the file"
        with open(self.filename, "a", newline="") as file:
            writer = self.get_writer(file)
            d = self.item_to_dict(item)
            writer.writerow(d)

    def get_writer(self, buff: TextIO):
        return csv.DictWriter(buff, fieldnames=self.get_headers(), **self.kwds_csv)

    def get_reader(self, buff: TextIO):
        return csv.DictReader(buff, fieldnames=self.get_headers(), **self.kwds_csv)

    def create(self, if_exists="raise"):
        "Create the repository file"
        if if_exists == "raise" and self.filename.exists():
            raise FileExistsError(f"Repository file {self.filename} already exists.")
        with open(self.filename, "w", newline="") as file:
            writer = self.get_writer(file)
            writer.writeheader()

    def data_to_item(self, data: dict) -> Item:
        # None values are converted to empty strings
        # Therefore we interpret empty strings as None
        data = {key: None if val == "" else val for key, val in data.items()}
        return super().data_to_item(data)




from copy import copy
from logging import Handler
import logging
from redbird.base import BaseRepo

class RepoHandler(Handler):
    """Log handler that writes log records to a repository

    Useful for cases where the log records need to be read
    programmatically.

    Parameters
    ----------
    repo : BaseRepo
        Repository where the log records are written
    **kwargs : dict
        Keyword arguments passed to logging.Handler
        init
    """

    def __init__(self, repo:BaseRepo, **kwargs):
        self.repo = repo
        super().__init__(**kwargs)

    def emit(self, record:logging.LogRecord):
        "Log the log record"
        record = copy(record)
        msg = self.format(record)
        if isinstance(msg, (dict, logging.LogRecord)):
            # Formatting returned the log record with formatted message
            record = msg
        else:
            # Formatting returned the message as string
            record.formatted_message = msg

        self.write(vars(record))

    def write(self, record:dict):
        "Write a log record to the repository"
        self.repo.add(record)





import logging
from typing import Optional
from pydantic import BaseModel

import pytest

from redbird.logging import RepoHandler
from redbird.repos import MemoryRepo
from redbird.base import BaseRepo

class LogRecord(BaseModel):
    name: str
    msg: str
    levelname: str
    levelno: int
    pathname: str
    filename: str
    module: str
    exc_info: Optional[tuple]
    exc_text: Optional[str]
    stack_info: Optional[tuple]
    lineno: int
    funcName: str
    created: float
    msecs: float
    relativeCreated: float
    thread: int
    threadName: str
    processName: str
    process: int

    message: str
    formatted_message: str


@pytest.fixture(scope="function", autouse=True)
def logger(request):
    name = __name__ + '.'.join(request.node.nodeid.split("::")[1:])
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    yield logger
    logger.handlers = []

@pytest.fixture(scope="function", autouse=True)
def handler(logger):
    handler = RepoHandler(repo=MemoryRepo())
    logger.addHandler(handler)
    return handler

def test_handler_attrs(logger):
    handler = RepoHandler(repo=MemoryRepo())
    logger.addHandler(handler)

    assert hasattr(handler, "repo")
    assert isinstance(handler.repo, BaseRepo)

def test_info(logger):
    repo = MemoryRepo()
    handler = RepoHandler(repo=repo)
    assert handler.repo is repo

    logger.addHandler(handler)

    logger.info("a log", extra={"myextra": "something"})

    records = repo.filter_by().all()
    assert isinstance(records, list)
    assert 1 == len(records)
    
    record = records[0]

    assert isinstance(record, dict)
    assert set(record.keys()) == {
        'name', 'msg', 'args', 'levelname', 'levelno', 
        'pathname', 'filename', 'module', 
        'exc_info', 'exc_text', 'stack_info', 'lineno', 'funcName', 
        'created', 'msecs', 'relativeCreated', 
        'thread', 'threadName', 'processName', 'process', 
        'myextra', 'message', 'formatted_message'
    }

    expected_values = {
        'levelname': 'INFO',
        'levelno': 20,
        'threadName': 'MainThread',
        'processName': 'MainProcess',
        
        'args': (),
        'exc_info': None,
        'exc_text': None,
        'stack_info': None,

        'message': 'a log',
        'myextra': 'something',
        'formatted_message': 'a log',
    }

    actual_values = {key: val for key, val in record.items() if key in expected_values}
    assert actual_values == expected_values

def test_model(logger):
    repo = MemoryRepo(model=LogRecord)
    handler = RepoHandler(repo=repo)
    assert handler.repo is repo

    logger.addHandler(handler)

    logger.info("a log", extra={"myextra": "something"})

    try:
        raise RuntimeError("Oops")
    except:
        logger.exception("Error occurred")

def test_filter(logger):
    repo = MemoryRepo()
    handler = RepoHandler(repo=repo)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)

    logger.info("an info", extra={"myextra": "something"})
    logger.debug("a debug", extra={"myextra": "something"})
    logger.warning("a warning", extra={"myextra": "something"})
    assert repo.filter_by().count() == 3

    records = repo.filter_by(levelname="INFO").all()
    assert len(records) == 1
    assert records[0]['levelname'] == 'INFO'

def test_set_level(logger):
    repo = MemoryRepo()
    handler = RepoHandler(repo=repo)
    handler.setLevel(logging.WARNING)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)

    logger.info("an info", extra={"myextra": "something"})
    logger.debug("a debug", extra={"myextra": "something"})
    logger.warning("a warning", extra={"myextra": "something"})

    records = repo.filter_by().all()
    assert len(records) == 1
    assert records[0]['levelname'] == 'WARNING'

def test_record_formatter(logger):
    # Test formatter that returns the LogRecord itself
    class DummyFormatter(logging.Formatter):
        def format(self, record):
            record.my_message = "Message: " + super().format(record)
            return record
    
    repo = MemoryRepo()
    handler = RepoHandler(repo=repo)
    handler.formatter = DummyFormatter()
    handler.setLevel(logging.WARNING)
    logger.addHandler(handler)

    logger.warning("an info")

    actual_record = repo.filter_by().first()
    assert actual_record['my_message'] == "Message: an info"
    assert 'formatted_message' not in actual_record



from setuptools import setup, find_packages
import versioneer

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="redbird",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Mikael Koli",
    author_email="koli.mikael@gmail.com",
    url="https://github.com/Miksus/red-bird.git",
    packages=find_packages(),
    description="Repository Patterns for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',

        "Intended Audience :: Developers",

        'Topic :: Database',
    ],
    include_package_data=True, # for MANIFEST.in
    python_requires='>=3.7.0',
    extras_require={
        "full": ["sqlalchemy", "pymongo", "requests", "pydantic-sqlalchemy"],
        "full-test": [
            "sqlalchemy", "pymongo", "requests", "pydantic-sqlalchemy",
            "pytest", "python-dotenv", "responses", "mongomock"
        ],
        "sql": ["sqlalchemy", "pydantic-sqlalchemy"],
        "mongodb": ['pymongo'],
        "rest": ["requests"],
    },
    install_requires = ["pydantic", "typing_extensions"],
)




from .base import BaseRepo, BaseResult
from . import _version
__version__ = _version.get_versions()['version']

def create_repo(model=None, model_orm=None, repo_name="MemoryRepo"):
    registry[repo_name](model=model, model_orm=model_orm)




import pytest
from pathlib import Path
import os
from dotenv import load_dotenv

def get_node_id(request):
    components = request.node.nodeid.split("::")
    filename = components[0]
    test_class = components[1] if len(components) == 3 else None
    test_func_with_params = components[-1]
    test_func = test_func_with_params.split('[')[0]

    filename = filename.replace(".py", "").replace("/", "-")
    if test_class:
        return f'{filename}-{test_class}-{test_func_with_params}'
    else:
        return f'{filename}-{test_func_with_params}'

@pytest.fixture(scope="function")
def mongo_uri():
    load_dotenv()
    pytest.importorskip("pymongo")
    if "MONGO_CONN" not in os.environ:
        pytest.skip()
    return os.environ["MONGO_CONN"]

@pytest.fixture(scope="function")
def collection(request, mongo_client):
    
    db_name = "pytest"
    col_name = get_node_id(request)

    collection = mongo_client[db_name][col_name]

    # Empty the collection
    collection.delete_many({})

    yield collection



from textwrap import dedent
from typing import Optional

import pytest

from redbird.repos import JSONDirectoryRepo
from pydantic import BaseModel

class Item(BaseModel):
    id: str
    name: str
    age: Optional[int]

def sort_items(items, repo, field="id"):
    return list(sorted(items, key=lambda x: repo.get_field_value(x, field)))

def test_missing_id(tmpdir):
    with pytest.raises(ValueError):
        repo = JSONDirectoryRepo(path=tmpdir, model=Item)

def test_filecontent(tmpdir):
    repo = JSONDirectoryRepo(path=tmpdir, model=Item, id_field="id")

    repo.add(Item(id="1111", name="Jack", age=20))
    repo.add(Item(id="2222", name="John"))
    repo.add(Item(id="3333", name="James", age=30))

    content = (tmpdir / "1111.json").read_text(encoding="UTF-8")
    assert content == '{"id": "1111", "name": "Jack", "age": 20}'

    repo.filter_by(id="1111").update(age=40)
    content = (tmpdir / "1111.json").read_text(encoding="UTF-8")
    assert content == '{"id": "1111", "name": "Jack", "age": 40}'

    repo.filter_by(id="2222").delete()
    assert not (tmpdir / "2222.json").exists()

    repo.filter_by().delete()
    assert not (tmpdir / "1111.json").exists()
    assert not (tmpdir / "3333.json").exists()


def test_kwds_json(tmpdir):
    repo = JSONDirectoryRepo(path=tmpdir, model=Item, id_field="id", kwds_json_dump={'indent': 4, 'sort_keys': True})

    repo.add(Item(id="1111", name="Jack", age=20))

    content = (tmpdir / "1111.json").read_text(encoding="UTF-8")
    assert content == dedent('''
    {
        "age": 20,
        "id": "1111",
        "name": "Jack"
    }
    ''')[1:-1]

    # Testing reading
    assert repo['1111'] == Item(id="1111", name="Jack", age=20)

def test_dict_operations(tmpdir):
    repo = JSONDirectoryRepo(path=tmpdir, id_field="id")

    # Insert
    repo.add(dict(id="a", name="Jack", age=20))
    repo.add(dict(id="b", name="Jack"))
    repo.add(dict(id="c", name="James", age=30))

    # Read
    actual = repo.filter_by().all()
    expected = [
        dict(id="a", name="Jack", age=20),
        dict(id="b", name="Jack"),
        dict(id="c", name="James", age=30)
    ]
    assert sort_items(actual, repo) == sort_items(expected, repo)

    actual = repo.filter_by(name="Jack").all() 
    expected = [
        dict(id="a", name="Jack", age=20),
        dict(id="b", name="Jack")
    ]
    assert sort_items(actual, repo) == sort_items(expected, repo)

    actual = repo.filter_by(age=30).all() 
    expected = [
        dict(id="c", name="James", age=30),
    ]
    assert sort_items(actual, repo) == sort_items(expected, repo)

    assert repo.filter_by(age="30").all() == []

    # Update
    repo.filter_by(name="Jack").update(age=50)
    actual = repo.filter_by().all()
    expected = [
        dict(id="a", name="Jack", age=50),
        dict(id="b", name="Jack", age=50),
        dict(id="c", name="James", age=30)
    ]
    assert sort_items(actual, repo) == sort_items(expected, repo)

    # Delete
    repo.filter_by(name="Jack").delete()
    assert repo.filter_by().all() == [
        #dict(id="a", name="Jack", age=50),
        #dict(id="b", name="Jack", age=50),
        dict(id="c", name="James", age=30)
    ]



import configparser
from typing import Optional
import re, os
import json
from dotenv import load_dotenv

import pytest
import responses
import requests
from redbird.repos.csv import CSVFileRepo
from redbird.repos.json import JSONDirectoryRepo
from redbird.repos.rest import RESTRepo
from redbird.repos.sqlalchemy import SQLRepo
from redbird.repos.memory import MemoryRepo
from redbird.repos.mongo import MongoRepo
from redbird.oper import in_, skip, between, greater_equal, greater_than, less_equal, less_than, not_equal

from pydantic import BaseModel, Field


TEST_CASES = [
    pytest.param("memory"),
    pytest.param("memory-dict"),
    pytest.param("sql-dict"),
    pytest.param("sql-pydantic"),
    pytest.param("sql-orm"),
    pytest.param("sql-pydantic-orm"),
    pytest.param("mongo"),
    pytest.param("mongo-mock"),
    pytest.param("http-rest"),
    pytest.param("csv"),
    pytest.param("json-dir"),
]


def sort_items(items, repo, field="id"):
    return list(sorted(items, key=lambda x: repo.get_field_value(x, field)))

# ------------------------
# ACTUAL TESTS
# ------------------------

class RepoTests:

    def populate(self, repo, items=None):
        if items is None:
            items = [
                dict(id="a", name="Jack", age=20),
                dict(id="b", name="John", age=30),
                dict(id="c", name="James", age=30),
                dict(id="d", name="Johnny", age=30),
                dict(id="e", name="Jesse", age=40),
            ]
        for d in items:
            item = repo.to_item(d)
            repo.add(item)

@pytest.mark.parametrize(
    'repo',
    TEST_CASES,
    indirect=True
)
class TestAPI:

    def test_session(self, repo):
        repo.session.remove()

    def test_has_attributes(self, repo):
        # Test given attrs are found
        attrs = [
            "session", "model",
            "add", "insert", "update", "upsert", "delete",
            "filter_by",
        ]
        for attr in attrs:
            getattr(repo, attr)

    def test_has_attributes_result(self, repo):
        attrs = [
            "query_", "repo", "query",
            "first", "all", "limit", "last",
            "update", "delete",
            "count",
        ]
        filter = repo.filter_by()
        for attr in attrs:
            getattr(filter, attr)


@pytest.mark.parametrize(
    'repo',
    TEST_CASES,
    indirect=True
)
class TestPopulated(RepoTests):

    def populate(self, repo, items=None):
        if items is None:
            items = [
                dict(id="a", name="Jack", age=20),
                dict(id="b", name="John", age=30),
                dict(id="c", name="James", age=30),
                dict(id="d", name="Johnny", age=30),
                dict(id="e", name="Jesse", age=40),
            ]
        for d in items:
            item = repo.to_item(d)
            repo.add(item)

    def test_filter_by_first(self, repo):
        self.populate(repo)
        Item = repo.model
        if repo.ordered:
            assert repo.filter_by(age=30).first() == Item(id="b", name="John", age=30)
        else:
            item = repo.filter_by(age=30).first()
            assert repo.get_field_value(item, 'age') == 30

    def test_filter_by_last(self, repo):
        self.populate(repo)
        Item = repo.model
        if repo.ordered:
            assert repo.filter_by(age=30).last() == Item(id="d", name="Johnny", age=30)
        else:
            item = repo.filter_by(age=30).last()
            assert repo.get_field_value(item, 'age') == 30

    def test_filter_by_limit(self, repo):
        self.populate(repo)
        Item = repo.model

        actual = repo.filter_by(age=30).limit(2)
        expected = [
                Item(id="b", name="John", age=30),
                Item(id="c", name="James", age=30),
            ]
        if not repo.ordered:
            # We don't know which are the top 2
            assert isinstance(actual, list)
            for item in actual:
                assert repo.get_field_value(item, "age") == 30
            assert len(actual) == 2
        else:
            assert actual == expected

    def test_filter_by_all(self, repo):
        self.populate(repo)
        Item = repo.model
        
        actual = repo.filter_by(age=30).all()
        expected = [
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
        ]
        
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_filter_by_iter(self, repo):
        self.populate(repo)
        Item = repo.model
        
        actual = list(repo.filter_by(age=30))
        expected = [
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
        ]
        
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_filter_by_update(self, repo):
        self.populate(repo)
        Item = repo.model

        repo.filter_by(age=30).update(name="Something")

        actual = repo.filter_by().all()
        expected = [
            Item(id="a", name="Jack", age=20),
            Item(id="b", name="Something", age=30),
            Item(id="c", name="Something", age=30),
            Item(id="d", name="Something", age=30),
            Item(id="e", name="Jesse", age=40),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_filter_by_replace(self, repo):
        self.populate(repo)
        Item = repo.model

        repo.filter_by(id="b").replace(id="b", name="Something")

        expected = [
            Item(id="a", name="Jack", age=20),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
            Item(id="b", name="Something", age=None),
        ]
        if repo.model == dict and not isinstance(repo, SQLRepo):
            # In dict repositories the not given fields don't exists
            # (exception: structured data stores)
            expected = [
                Item(id="a", name="Jack", age=20),
                Item(id="c", name="James", age=30),
                Item(id="d", name="Johnny", age=30),
                Item(id="e", name="Jesse", age=40),
                Item(id="b", name="Something"),
            ]

        actual = repo.filter_by().all()
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)

        assert actual == expected

    def test_filter_by_replace_dict(self, repo):
        self.populate(repo)
        Item = repo.model

        repo.filter_by(id="b").replace({"id": "b", "name": "Something"})
        
        expected = [
            Item(id="a", name="Jack", age=20),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
            Item(id="b", name="Something", age=None),
        ]
        if repo.model == dict and not isinstance(repo, SQLRepo):
            # In dict repositories the not given fields don't exists
            # (exception: structured data stores)
            expected = [
                Item(id="a", name="Jack", age=20),
                Item(id="c", name="James", age=30),
                Item(id="d", name="Johnny", age=30),
                Item(id="e", name="Jesse", age=40),
                Item(id="b", name="Something"),
            ]

        actual = repo.filter_by().all()
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)

        assert actual == expected

    def test_filter_by_delete(self, repo):
        self.populate(repo)
        Item = repo.model

        repo.filter_by(age=30).delete()

        actual = repo.filter_by().all()
        expected = [
            Item(id="a", name="Jack", age=20),
            #Item(id="b", name="John", age=30),
            #Item(id="c", name="James", age=30),
            #Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_filter_by_count(self, repo):
        self.populate(repo)
        Item = repo.model
        assert repo.filter_by(age=30).count() == 3

    def test_filter_by_less_than(self, repo):
        self.populate(repo)

        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")

        Item = repo.model
        actual = repo.filter_by(age=less_than(40)).all() 
        expected = [
            Item(id="a", name="Jack", age=20),
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            #Item(id="e", name="Jesse", age=40),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_getitem(self, repo):
        self.populate(repo)
        Item = repo.model

        assert repo["b"] == Item(id="b", name="John", age=30)

    def test_getitem_missing(self, repo):
        self.populate(repo)
        with pytest.raises(KeyError):
            repo["not_found"]

    def test_delitem(self, repo):
        self.populate(repo)
        Item = repo.model

        del repo["b"]


        actual = repo.filter_by().all() 
        expected = [
            Item(id="a", name="Jack", age=20),
            #Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    @pytest.mark.skip(reason="This is a minor issue")
    def test_delitem_missing(self, repo):
        self.populate(repo)
        with pytest.raises(KeyError):
            del repo["not_found"]

    def test_setitem(self, repo):
        self.populate(repo)
        Item = repo.model

        repo["d"] = {"name": "Johnny boy"}

        actual = repo.filter_by().all() 
        expected = [
            Item(id="a", name="Jack", age=20),
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny boy", age=30),
            Item(id="e", name="Jesse", age=40),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    @pytest.mark.skip(reason="This is a minor issue")
    def test_setitem_missing(self, repo):
        self.populate(repo)
        with pytest.raises(KeyError):
            repo["not_found"] = {"name": "something"}

@pytest.mark.parametrize(
    'repo',
    TEST_CASES,
    indirect=True
)
class TestFilteringOperations(RepoTests):

    def populate(self, repo, items=None):
        if items is None:
            items = [
                dict(id="a", name="Jack", age=20),
                dict(id="b", name="John", age=30),
                dict(id="c", name="James", age=30),
                dict(id="d", name="Johnny", age=30),
                dict(id="e", name="Jesse", age=40),
                dict(id="f", name="Jim", age=41),
            ]
        for d in items:
            item = repo.to_item(d)
            repo.add(item)

    def test_greater_than(self, repo):
        self.populate(repo)

        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")

        Item = repo.model

        actual = repo.filter_by(age=greater_than(20)).all()
        expected = [
            #Item(id="a", name="Jack", age=20),
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
            Item(id="f", name="Jim", age=41),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_less_than(self, repo):
        self.populate(repo)

        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")
    
        Item = repo.model
        actual = repo.filter_by(age=less_than(30)).all()
        expected = [
            Item(id="a", name="Jack", age=20),
            #Item(id="b", name="John", age=30),
            #Item(id="c", name="James", age=30),
            #Item(id="d", name="Johnny", age=30),
            #Item(id="e", name="Jesse", age=40),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_greater_equal(self, repo):
        self.populate(repo)
        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")
        Item = repo.model
        actual = repo.filter_by(age=greater_equal(30)).all()
        expected = [
            #Item(id="a", name="Jack", age=20),
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
            Item(id="f", name="Jim", age=41),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_less_equal(self, repo):
        self.populate(repo)

        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")

        Item = repo.model
        actual = repo.filter_by(age=less_equal(30)).all()
        expected = [
            Item(id="a", name="Jack", age=20),
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            #Item(id="e", name="Jesse", age=40),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_not_equal(self, repo):
        self.populate(repo)

        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")

        Item = repo.model
        actual = repo.filter_by(age=not_equal(30)).all()
        expected = [
            Item(id="a", name="Jack", age=20),
            #Item(id="b", name="John", age=30),
            #Item(id="c", name="James", age=30),
            #Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
            Item(id="f", name="Jim", age=41),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_between(self, repo):
        self.populate(repo)
        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")
        Item = repo.model
        actual = repo.filter_by(age=between(30, 40)).all() 
        expected = [
            #Item(id="a", name="Jack", age=20),
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
            #Item(id="e", name="Jim", age=41),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_in(self, repo):
        self.populate(repo)
        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")
        Item = repo.model
        actual = repo.filter_by(age=in_([20, 41])).all() 
        expected = [
            Item(id="a", name="Jack", age=20),
            #Item(id="b", name="John", age=30),
            #Item(id="c", name="James", age=30),
            #Item(id="d", name="Johnny", age=30),
            #Item(id="e", name="Jesse", age=40),
            Item(id="f", name="Jim", age=41),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_skip(self, repo):
        self.populate(repo)
        if isinstance(repo, RESTRepo):
            pytest.xfail("RESTRepo does not support operations (yet)")
        Item = repo.model
        actual = repo.filter_by(age=skip).all()
        expected = [
            Item(id="a", name="Jack", age=20),
            Item(id="b", name="John", age=30),
            Item(id="c", name="James", age=30),
            Item(id="d", name="Johnny", age=30),
            Item(id="e", name="Jesse", age=40),
            Item(id="f", name="Jim", age=41),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

@pytest.mark.parametrize(
    'repo',
    TEST_CASES,
    indirect=True
)
class TestEmpty:

    def test_add(self, repo):
        Item = repo.model
        
        assert repo.filter_by().all() == []

        repo.add(Item(id="a", name="Jack", age=20))
        assert repo.filter_by().all() == [Item(id="a", name="Jack", age=20)]

        repo.add(Item(id="b", name="John", age=30))
        
        actual = repo.filter_by().all()
        expected = [
            Item(id="a", name="Jack", age=20),
            Item(id="b", name="John", age=30)
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_update(self, repo):
        Item = repo.model

        repo.add(Item(id="a", name="Jack", age=20))
        repo.add(Item(id="b", name="John", age=30))

        repo.update(Item(id="a", name="Max"))
        actual = repo.filter_by().all()

        actual = repo.filter_by().all()
        expected = [
            Item(id="a", name="Max", age=20),
            Item(id="b", name="John", age=30),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

    def test_replace(self, repo):
        Item = repo.model

        repo.add(Item(id="a", name="Jack", age=20))
        repo.add(Item(id="b", name="John", age=30))

        repo.replace(Item(id="a", name="Max"))

        if repo.model == dict and not isinstance(repo, SQLRepo):
            expected = [
                Item(id="a", name="Max"),
                Item(id="b", name="John", age=30),
            ]
        else:
            expected = [
                Item(id="a", name="Max", age=None),
                Item(id="b", name="John", age=30),
            ]

        actual = repo.filter_by().all()
        actual = sort_items(actual, repo)
        expected = sort_items(expected, repo)
        assert actual == expected

    def test_add_exist(self, repo):
        Item = repo.model
    
        repo.add(Item(id="a", name="Jack", age=20))
        with pytest.raises(Exception):
            repo.add(Item(id="a", name="John", age=30))

    def test_add_exist_ignore(self, repo):
        Item = repo.model
    
        repo.add(Item(id="a", name="Jack", age=20), if_exists="ignore")
        repo.add(Item(id="a", name="John", age=30), if_exists="ignore")
        assert repo.filter_by().all() == [
            Item(id="a", name="Jack", age=20),
        ]

    def test_add_exist_update(self, repo):
        Item = repo.model
    
        repo.add(Item(id="a", name="Jack", age=20), if_exists="update")
        assert repo.filter_by().all() == [
            Item(id="a", name="Jack", age=20),
        ]

        repo.add(Item(id="a", name="John", age=30), if_exists="update")
        assert repo.filter_by().all() == [
            Item(id="a", name="John", age=30),
        ]

    def test_get_by(self, repo):
        Item = repo.model

        repo.add(Item(id="a", name="Max", age=20))
        repo.add(Item(id="b", name="John", age=30))

        # asserting
        actual = list(repo)
        expected =[
            Item(id="a", name="Max", age=20),
            Item(id="b", name="John", age=30),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

        repo.get_by("a").update(age=40)

        actual = list(repo)
        expected = [
            Item(id="a", name="Max", age=40),
            Item(id="b", name="John", age=30),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected

        repo.get_by("b").delete()
        assert list(repo) == [
            Item(id="a", name="Max", age=40),
        ]

@pytest.mark.parametrize(
    'repo',
    TEST_CASES,
    indirect=True
)
class TestMalformedData(RepoTests):

    def test_filter_raise(self, repo):

        class MalformedItem(BaseModel):
            id: str
            name: int
            age: int

        self.populate(repo, [
            {"id": "a", "name": 1, "age": 20},
            {"id": "b", "name": "Jack", "age": 20},
            {"id": "c", "name": "James", "age": 30},
            {"id": "d", "name": 2, "age": 30},
            {"id": "e", "name": 3, "age": 30},
        ])
        repo.model = MalformedItem

        with pytest.raises(ValueError):
            repo.filter_by().all()

    def test_filter_warn(self, repo):

        class MalformedItem(BaseModel):
            id: str
            name: int
            age: int

        self.populate(repo, [
            {"id": "a", "name": 1, "age": 20},
            {"id": "b", "name": "Jack", "age": 20},
            {"id": "c", "name": "James", "age": 30},
            {"id": "d", "name": 2, "age": 30},
            {"id": "e", "name": 3, "age": 30},
        ])
        repo.model = MalformedItem
        repo.errors_query = "warn"

        with pytest.warns(UserWarning):
            repo.filter_by().all()

    def test_filter_discard(self, repo):

        class MalformedItem(BaseModel):
            id: str
            name: int
            age: int

        self.populate(repo, [
            {"id": "a", "name": 1, "age": 20},
            {"id": "b", "name": "Jack", "age": 20},
            {"id": "c", "name": "James", "age": 30},
            {"id": "d", "name": 2, "age": 30},
            {"id": "e", "name": 3, "age": 30},
        ])
        repo.model = MalformedItem
        repo.errors_query = "discard"

        Item = MalformedItem
        actual = repo.filter_by().all() 
        expected = [
            Item(id="a", name=1, age=20),
            Item(id="d", name=2, age=30),
            Item(id="e", name=3, age=30),
        ]
        if not repo.ordered:
            actual = sort_items(actual, repo)
            expected = sort_items(expected, repo)
        assert actual == expected



from typing import TYPE_CHECKING, Any, Optional, Type
from pydantic import BaseModel, Field, PrivateAttr
from redbird import BaseRepo, BaseResult
from redbird.templates import TemplateRepo
from redbird.exc import KeyFoundError


from redbird.oper import Between, In, Operation, skip
from redbird.utils.deprecate import deprecated

if TYPE_CHECKING:
    from sqlalchemy.engine import Engine



class SQLRepo(TemplateRepo):
    """SQL Repository

    Parameters
    ----------
    conn_string : str, optional
        Connection string to the database. 
        Pass either conn_string, engine or session if
        model_orm is not defined.
    engine : sqlalchemy.engine.Engine, optional
        SQLAlchemy engine to connect the database. 
        Pass either conn_string, engine or session if
        model_orm is not defined.
    model : Type
        Class of an item in the repository.
        Commonly dict or subclass of Pydantic
        BaseModel. By default dict
    id_field : str, optional
        Attribute or key that identifies each item
        in the repository.
    field_access : {'attr', 'key'}, optional
        How to access a field in an item. Either
        by attribute ('attr') or key ('item').
        By default guessed from the model.
    query : Type, optional
        Query model of the repository.
    errors_query : {'raise', 'warn', 'discard'}
        Whether to raise an exception, warn or discard
        the item in case of validation error in 
        converting data to the item model from
        the repository. By default raise 
    model_orm : Type of Base, optional
        Subclass of SQL Alchemy representation of the item.
        This is the class that is operated behind the scenes.
    table : str, optional
        Table name where the items lies. Should only be given
        if no model_orm specified.
    session : sqlalchemy.session.Session
        Connection session to the database.
        Pass either conn_string, engine or session if
        model_orm is not defined.

    Examples
    --------

    .. code-block:: python

        from sqlalchemy import create_engine
        from redbird.repos import SQLRepo

        engine = create_engine('sqlite://')
        repo = SQLRepo(engine=engine, table="my_table")

    You may also supply a session:

    .. code-block:: python

        from sqlalchemy import create_engine
        from sqlalchemy.orm import Session
        from redbird.repos import SQLRepo

        engine = create_engine('sqlite://')
        session = Session(engine)
        repo = SQLRepo(session=session, table="my_table")

    Using ORM model:

    .. code-block:: python

        from sqlalchemy.orm import declarative_base
        
        Base = declarative_base()

        class Car(Base):
            __tablename__ = 'my_table'
            color = Column(String, primary_key=True)
            car_type = Column(String)
            milage = Column(Integer)

        from sqlalchemy import create_engine
        from redbird.repos import SQLRepo

        engine = create_engine('sqlite://')
        repo = SQLRepo(model_orm=Car, engine=engine)

    Using ORM model and reflect Pydantic Model:

    .. code-block:: python

        from sqlalchemy.orm import declarative_base
        
        Base = declarative_base()

        class Car(Base):
            __tablename__ = 'my_table'
            color = Column(String, primary_key=True)
            car_type = Column(String)
            milage = Column(Integer)

        from sqlalchemy import create_engine
        from redbird.repos import SQLRepo

        engine = create_engine('sqlite://')
        repo = SQLRepo(model_orm=Car, reflect_model=True, engine=engine)

    Using ORM model and Pydantic Model:

    .. code-block:: python

        from pydantic import BaseModel
        from sqlalchemy.orm import declarative_base
        
        Base = declarative_base()

        class CarORM(Base):
            __tablename__ = 'my_table'
            color = Column(String, primary_key=True)
            car_type = Column(String)
            milage = Column(Integer)

        class Car(BaseModel):
            id: str
            name: str
            age: int

        from sqlalchemy import create_engine
        from redbird.repos import SQLRepo

        engine = create_engine('sqlite://')
        repo = SQLRepo(model=Car, model_orm=CarORM, engine=engine)
    """

    model_orm: Optional[Any]
    table: Optional[str]
    session: Any
    engine: Optional[Any]
    autocommit: bool = Field(default=True, description="Whether to automatically commit the writes (create, delete, update)")

    ordered: bool = Field(default=True, const=True)
    _Base = PrivateAttr()

    @classmethod
    @deprecated("Please use normal init instead.")
    def from_engine(cls, *args, engine, **kwargs):
        kwargs["session"] = cls.create_scoped_session(engine)
        return cls(*args, **kwargs)

    @classmethod
    @deprecated("Please use normal init instead.")
    def from_connection_string(cls, *args, conn_string, **kwargs):
        from sqlalchemy import create_engine
        return cls.from_engine(*args, engine=create_engine(conn_string), **kwargs)

    def __init__(self, *args, reflect_model=False, conn_string=None, engine=None, session=None, if_missing="raise", **kwargs):
        from sqlalchemy import create_engine
        from sqlalchemy.ext.automap import automap_base
        from sqlalchemy.exc import NoSuchTableError

        # Determine connection
        if conn_string is not None:
            engine = create_engine(conn_string)
        if engine is not None:
            session = self.create_scoped_session(engine)

        # Create model_orm/model
        if "model_orm" not in kwargs:
            if session is None:
                raise TypeError(
                    "Connection cannot be determined. "
                    "Consider using method 'from_connection_string' "
                    "and pass connection string as conn_string"
                )
            table = kwargs["table"]
            table_exists = session.get_bind().has_table(table)
            if not table_exists:
                if if_missing == "raise":
                    raise NoSuchTableError(f"Table {table} is missing. Create the table or pass if_missing='create'")
                elif if_missing == "create":
                    self._create_table(session, kwargs['model'], name=table, primary_column=kwargs.get('id_field'))

            self._Base = automap_base()
            self._Base.prepare(engine=session.get_bind(), reflect=True)
            try:
                model_orm = self._Base.classes[table]
            except KeyError as exc:
                raise KeyError(f"Cannot automap table '{table}'. Perhaps table missing primary key?") from exc
            kwargs["model_orm"] = model_orm
        if reflect_model:
            kwargs["model"] = self.orm_model_to_pydantic(kwargs["model_orm"])
        super().__init__(*args, session=session, **kwargs)

    def insert(self, item):
        from sqlalchemy.exc import IntegrityError
        row = self.item_to_data(item)
        self.session.add(row)
        if self.autocommit:
            try:
                self.session.commit()
            except IntegrityError as exc:
                self.session.rollback()
                raise KeyFoundError(f"Item {self.get_field_value(item, self.id_field)} is already in the table.") from exc

    def upsert(self, item):
        row = self.item_to_data(item)

        self.session.merge(row)
        self.session.commit()

    def data_to_item(self, item_orm):
        # Turn ORM item to Pydantic item
        if hasattr(self.model, "from_orm") and self.model.Config.orm_mode:
            # Use Pydantic methods
             return self.model.from_orm(item_orm)
        else:
            # Turn ORM to model using mapping
            d = vars(item_orm)
            d.pop("_sa_instance_state", None)
            return self.model(**d)

    def item_to_data(self, item:BaseModel):
        # Turn Pydantic item to ORM item
        return self.model_orm(**self.item_to_dict(item))

    def orm_model_to_pydantic(self, model):
        # Turn SQLAlchemy BaseModel to Pydantic BaseModel
        from pydantic_sqlalchemy import sqlalchemy_to_pydantic
        return sqlalchemy_to_pydantic(model)

    @staticmethod
    def create_scoped_session(engine):
        from sqlalchemy.orm import sessionmaker, scoped_session
        Session = sessionmaker(bind=engine)
        return scoped_session(Session)

    def item_to_dict(self, item):
        if isinstance(item, dict):
            return item
        elif hasattr(item, "dict"):
            # Is pydantic
            return item.dict(exclude_unset=True)
        else:
            d = vars(item)
            d.pop("_sa_instance_state", None)
            return d

    def create(self):
        "Create the database and table"
        self.model_orm.__table__.create(bind=self.session.bind)

    def query_data(self, query):
        for data in self._filter_orm(query):
            yield data

    def query_data_first(self, query):
        item = self._filter_orm(query).first()
        if item is not None:
            return self.data_to_item(item)

    def query_update(self, query, values):
        self._filter_orm(query).update(values)
        if self.autocommit:
            self.session.commit()

    def query_delete(self, query):
        self._filter_orm(query).delete()
        if self.autocommit:
            self.session.commit()

    def query_count(self, query):
        return self._filter_orm(query).count()

    def _filter_orm(self, query):
        session = self.session
        return session.query(self.model_orm).filter(query)

    def format_query(self, oper: dict):
        from sqlalchemy import Column, orm, true
        stmt = true()
        for column_name, oper_or_value in oper.items():
            column = getattr(self.model_orm, column_name) if self.model_orm is not None else Column(column_name)
            if isinstance(oper_or_value, Operation):
                oper = oper_or_value
                if isinstance(oper, Between):
                    sql_oper = column.between(oper.start, oper.end)
                elif isinstance(oper, In):
                    sql_oper = column.in_(oper.value)
                elif oper is skip:
                    continue
                elif hasattr(oper, "__py_magic__"):
                    magic = oper.__py_magic__
                    oper_method = getattr(column, magic)

                    # Here we form the SQLAlchemy operation, ie.: column("mycol") >= 5
                    sql_oper = oper_method(oper.value)
                else:
                    raise NotImplementedError(f"Not implemented operator: {oper}")
            else:
                value = oper_or_value
                sql_oper = column == value
            stmt &= sql_oper
        return stmt

    def _create_table(self, session, model, name, primary_column=None):
        from sqlalchemy import Table, Column, MetaData
        from sqlalchemy import String, Integer, Float, Boolean
        types = {
            str: String,
            int: Integer,
            float: Float,
            bool: Boolean
        }
        columns = [
            Column(name, types[field.type_], primary_key=name == primary_column)
            for name, field in model.__fields__.items()
        ]
        meta = MetaData()
        table = Table(name, meta, *columns)

        engine = session.get_bind()
        table.create(engine)